/***************************************************************************
                          cconnection.cpp  -  handles connections
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Po Jul 22 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cconnection.h"

#include <config.h>

#include <kmessagebox.h>
#include <klocale.h>
#include <kapplication.h>
#include <kstandarddirs.h>
#include <qdatetime.h>
#include <qtimer.h>

#include "kmuddy.h"
#include "ccmdprocessor.h"
#include "cconnectioninfo.h"
#include "cconnprefs.h"
#include "cprofiles.h"
#include "csession.h"
#include "csessionmanager.h"
#include "ctelnet.h"

#include "caliaslist.h"
#include "ctriggerlist.h"
#include "cactionlist.h"
#include "cgaugelist.h"
#include "cgrouplist.h"
#include "ctimerlist.h"
#include "cshortcutlist.h"
#include "cstatusvarlist.h"
#include "cvariablelist.h"
#include "cvartriggerlist.h"
#include "cwindowlist.h"

#include "ctranscript.h"

#include <map>

#include "dialogs/dlgprofilesettings.h"
#include "dialogs/dlgaliastrigger.h"
#include "dialogs/dlggauges.h"
#include "dialogs/dlgshortcuts.h"
#include "dialogs/dlgstatusvars.h"
#include "dialogs/dlgtimers.h"
#include "dialogs/dlgtimertick.h"
#include "dialogs/dlgvariables.h"
#include "dialogs/dlgvartriglist.h"
#include "dialogs/dlgwindows.h"

extern KMuddy *kmuddy;

#define AUTOSAVE_INTERVAL 5

// TODO: lists will go to a separate class - cListManager

/** this class holds private data of a cConnection object */
struct cConnectionPrivate {

  cConnectionInfo *ci;
  cConnPrefs *cp;

  QString caption;

  QString commands;
  // waiting for commands - no immediate sending
  bool waitCommands;

  int commandNum;

  QTimer *connecting;
  QTimer *saver;

  dlgProfileSettings *sdlg;

  bool destroying;

  bool connclosed;
  bool canCT;

  bool connInProgress, qconnInProgress, sendNothing;

  //login sequence and its iterator
  QStringList loginSeq;
  QStringList::iterator seqIter;

  //lists with data
  std::map<QString, cSaveableList *> dataLists;
  cVariableList *variablelist;
  cWindowList *windowlist;
};

cConnection::cConnection (int sess) : cActionBase ("connection", sess)
{
  d = new cConnectionPrivate;
  d->ci = 0;
  d->cp = 0;
  d->destroying = false;
  d->connecting = 0;
  d->connclosed = false;
  d->canCT = false;
  d->connInProgress = d->qconnInProgress = d->sendNothing = false;
  d->commands = QString::null;
  d->waitCommands = false;
  d->commandNum = 0;

  d->saver = new QTimer;
  connect (d->saver, SIGNAL (timeout ()), this, SLOT (saverTimeout ()));

  addEventHandler ("connected", 10, PT_NOTHING);
  addEventHandler ("save", 10, PT_NOTHING);
  addEventHandler ("disconnected", 100, PT_NOTHING);
  addEventHandler ("connection-failed", 100, PT_STRING);
  addEventHandler ("send-command", 50, PT_STRING);
  addEventHandler ("command-block", 50, PT_NOTHING);
  addEventHandler ("send-commands", 50, PT_NOTHING);
}

cConnection::~cConnection ()
{
  d->destroying = true;
  disconnect ();  //maybe we aren't connected, but disconnect() can handle it

  removeEventHandler ("connected");
  removeEventHandler ("disconnected");
  removeEventHandler ("connection-failed");
  removeEventHandler ("send-command");
  removeEventHandler ("command-block");
  removeEventHandler ("send-commands");

  delete d->ci;
  delete d->saver;
  delete d;
}

void cConnection::eventHandler (QString event, int)
{
  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return;

  if (event == "connected") {
    // TODO: better way of detecting quick/normal - other classes need it too
    // probably just some setting set upon creating objects (loader knows session type)

    d->connclosed = false;
    d->canCT = false;
    d->commands = QString::null;
    updateMenus ();
    kmuddy->enableConnectedOptions ();

    if (d->connInProgress) {
      d->ci->setIsQuick (false);
      //we can no longer reconnect
      //create a preferences object
      d->cp = new cConnPrefs (d->ci->name (), sess());
      //then load all preferences, constructing associated objects if needed
      d->cp->load ();
      loadLists ();
      invokeEvent ("load", sess());

      // set some variables - $character, $session
      cVariableList *vl = dynamic_cast<cVariableList *>(object ("variablelist"));
      cSession *s = dynamic_cast<cSession *>(object ("session"));
      QString character = d->ci->login();
      QString session = s->name();
      vl->set ("character", character);
      vl->set ("session", session);

      kmuddy->enableAdvancedOptions ();

      //update caption
      d->caption = ((d->ci->login().length()>0) ?
            (d->ci->login() + "@" + d->ci->name()) : d->ci->name());
      cSessionManager::self()->setSessionName (sess(), d->caption);

      // Here be autologging
      if(d->cp->autoAdvTranscript())
      {
        cTranscript *transcript = dynamic_cast<cTranscript *>(object ("transcript"));
        transcript->load();
        transcript->startAdvTranscript();
      }

      //everything's ready - schedule sending of login sequence if needed...
      if (!d->sendNothing)
      {
        d->loginSeq = d->ci->connStr();
        d->seqIter = d->loginSeq.begin();
        if ((d->loginSeq.count() > 1) || ((d->loginSeq.count() == 1) && (d->loginSeq[0].length() > 0)))
        {
          d->connecting = new QTimer;
          connect (d->connecting, SIGNAL (timeout()), this,
              SLOT (sendLoginAndPassword()));
          telnet->waitingForData();
          d->connecting->start (400);
        }
      }

      //also schedule automatic saving of settings and objects
      d->saver->start (AUTOSAVE_INTERVAL * 60 * 1000);
      (KApplication::kApplication())->processEvents ();

    } else
    {
      d->ci->setIsQuick (true);
      d->cp = NULL;
      kmuddy->disableAdvancedOptions ();
      invokeEvent ("load", sess());

  #ifdef HAVE_MXP
      //allow MXP auto-detection for quick connections
      telnet->setMXPAllowed (3);
  #endif

      //update application caption
      //update caption
      d->caption = d->ci->server();
      cSessionManager::self()->setSessionName (sess(), d->caption);
    }

    d->connInProgress = d->qconnInProgress = false;
  }
  else if (event == "disconnected") {
    disconnect ();
  }
  else if (event == "save") {
    save ();
  }
  else if (event == "command-block") {
    d->waitCommands = true;
  }
  else if (event == "send-commands") {
    sendCommands ();
  }
}

void cConnection::eventHandler (QString event, int, QString &par1, const QString &)
{
  if (event == "send-command") {
    addCommand (par1);
  }
  else if (event == "connection-failed") {
    QString error = i18n ("Could not establish connection.");
    error += "\n\n";
    error += i18n ("Reason: ");
    error += par1;
    KMessageBox::sorry (kmuddy, error, i18n ("Connect"));
  }
}

QString cConnection::actionHandler (QString action, int, QString &par1, const QString &)
{
  if (action == "send-command") {
    addCommand (par1);
  }

  return QString::null;
}

QString cConnection::actionHandler (QString action, int)
{
  if (action == "command-block") {
    d->waitCommands = true;
  }
  else if (action == "send-commands") {
    sendCommands ();
  }

  return QString::null;
}

cConnectionInfo *cConnection::getConnectionInfo () {
  return d->ci;
}

QString cConnection::getCaption () {
  return d->caption;
}

void cConnection::setConnectionClosed (bool canCloseTab) {
  d->connclosed = true;
  d->canCT = canCloseTab;
}

void cConnection::unsetConnectionClosed () {
  d->connclosed = false;
  d->canCT = false;
}

bool cConnection::connectionClosed () {
  return d->connclosed;
}

cSaveableList *cConnection::getList (const QString &name)
{
  if (d->dataLists.count (name))
    return d->dataLists[name];
  return 0;
}

void cConnection::showConnPrefsDialog ()
{
  if (d->cp == 0) //not a profile-based connection - do nothing
    return;

  //so first we have to create the dialog...
  d->sdlg = new dlgProfileSettings (kmuddy);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (d->sdlg, SIGNAL (okClicked()), this,
      SLOT (getSettingsFromDialog()));
  connect (d->sdlg, SIGNAL (applyClicked()), this,
      SLOT (getSettingsFromDialog()));

  //next we fill in its data
  putSettingsToDialog ();

  //dialog is ready - show it!
  d->sdlg->exec ();

  //further action is handled via slots issued by buttons, so that we only
  //have to destroy the dialog...
  delete d->sdlg;
}

void cConnection::getSettingsFromDialog ()
{
  cConnPrefs *cp = d->cp;
  dlgProfileSettings *sdlg = d->sdlg;

  cp->setAnsiColors (sdlg->useAnsi ());
  cp->setLimitTriggers (sdlg->limitTriggers ());
  cp->setLimitRepeater (sdlg->limitRepeater ());
  cp->setNegotiateOnStartup (sdlg->negotiateOnStartup());
  cp->setPromptLabel (sdlg->promptLabel ());
  cp->setLPMudStyle(sdlg->LPMudStyle());
  cp->setStatusPrompt(sdlg->statusPrompt());
  cp->setConsolePrompt(sdlg->consolePrompt());
  cp->setAutoAdvTranscript(sdlg->autoAdvTranscript());

  for (int i = 0; i < 10; i++)
    cp->setCmd (i, sdlg->movementCmd (i));
  cp->setQuit (sdlg->quitCmd ());
  cp->setScriptDir (sdlg->scriptDir ());
  cp->setWorkDir (sdlg->scriptWorkDir ());
  cp->setTranscriptDir (sdlg->transcriptDir ());

#ifdef HAVE_SOUND
  cp->setSoundDirs (sdlg->soundDirList());
  cp->setUseMSP (sdlg->useMSP());
  cp->setAlwaysMSP (sdlg->alwaysMSP());
  cp->setMidlineMSP (sdlg->midlineMSP());
#endif

#ifdef HAVE_MXP
  cp->setUseMXP (sdlg->MXPAllowed ());
  cp->setVarPrefix (sdlg->variablePrefix());
#endif  //HAVE_MXP

  //inform everyone ...
  invokeEvent ("settings-changed", sess());
}

void cConnection::putSettingsToDialog ()
{
  cConnPrefs *cp = d->cp;
  dlgProfileSettings *sdlg = d->sdlg;

  sdlg->setUseAnsi (cp->ansiColors ());
  sdlg->setLimitTriggers (cp->limitTriggers());
  sdlg->setLimitRepeater (cp->limitRepeater());
  sdlg->setNegotiateOnStartup (cp->negotiateOnStartup());
  sdlg->setPromptLabel(cp->promptLabel ());
  sdlg->setLPMudStyle(cp->LPMudStyle());
  sdlg->setStatusPrompt(cp->statusPrompt());
  sdlg->setConsolePrompt(cp->consolePrompt());
  sdlg->setAutoAdvTranscript(cp->autoAdvTranscript());

  for (int i = 0; i < 10; i++)
    sdlg->setMovementCmd (i, cp->cmd (i));
  sdlg->setQuitCmd (cp->quit ());
  sdlg->setScriptDir (cp->scriptDir ());
  sdlg->setScriptWorkDir (cp->workDir ());
  sdlg->setTranscriptDir (cp->transcriptDir ());

#ifdef HAVE_SOUND
  sdlg->setSoundDirList (cp->soundDirs());
  sdlg->setUseMSP (cp->useMSP());
  sdlg->setAlwaysMSP (cp->alwaysMSP());
  sdlg->setMidlineMSP (cp->midlineMSP());
#endif //HAVE_SOUND

#ifdef HAVE_MXP
  sdlg->setMXPAllowed (cp->useMXP ());
  sdlg->setVariablePrefix (cp->varPrefix());
#endif  //HAVE_MXP
}

bool cConnection::handleDisconnect ()
{
  if (!isConnected())
    return true;    //no longer connected
  //ask if he really wants to disconnect
  if (KMessageBox::questionYesNoCancel (kmuddy,
        i18n("Do you want to close connection to "+d->ci->name()+"?"),
        i18n("Disconnect")) == KMessageBox::Yes)
  {
    disconnect ();
    return true;
  }
  return false;
}

void cConnection::establishConnection (cConnectionInfo *cci,
    bool sendNothing)
{
  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return;

  if ((d->ci != cci) && (d->ci != 0))
  //old connection info remembered and it's not the current one
  {
    delete d->ci;
    d->ci = 0;
  }

  if (telnet->isConnected ())
    disconnect ();

  d->connInProgress = true;
  d->qconnInProgress = false;
  d->ci = cci;
  d->sendNothing = sendNothing;
  telnet->connectIt(cci->server(), cci->port ());
  // now we wait for event from telnet (standard connections only, offline ones
  // return immediately)
}

void cConnection::sendLoginAndPassword ()
{
  //This function is used to send login sequence (usually login name and
  //password) to the MUD.
  //Problem: it's not very good at detecting login prompt - name can be sent
  //too early, causing password to be sent too early as well...
  //I can only hope that the server will handle that properly.

  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return;
  static int wasData = false; //have some data already arrived?

  if (telnet->isConnected())
  {
    if (wasData)
    {
      if (telnet->newData() || (d->seqIter == d->loginSeq.begin()))
      //only send next command if something new has arrived (should be reply
      //to previous sent command). Also send it if it's the first command,
      //where the required data was received during previous call - used to
      //wait for intro screens and such).
      {
        //get next command from login sequence
        QString command = *d->seqIter;
        QString cmd = "", cmddisp = "";

        //expand $name and $password
        int len = command.length();
        QString varname = QString::null;

        bool expanded = false;
        bool invar = false;
        for (int i = 0; i < len; i++)
        {
          if (invar)
          {
            varname += command[i];
            if (varname == "$name")
            {
              varname = "";
              invar = false;
              cmd += d->ci->login();
              cmddisp += d->ci->login();
              expanded = true;
            }
            if (varname == "$password")
            {
              varname = "";
              invar = false;
              cmd += d->ci->password();
              cmddisp += i18n("(password is hidden)");
              expanded = true;
            }
          }
          else
            if (command[i].latin1() == '$')
            {
              invar = true;
              varname = "$";
            }
            else
            {
              cmd += command[i];
              cmddisp += command[i];
            }
        }

        //send resulting command and display notification
        //lines containing the name and password are sent as-is, others are sent to the
        //command processor - hence aliases/script names/... get expanded
        if (expanded)
        {
          telnet->sendData(cmd);
          telnet->sendData("\n");
          invokeEvent ("command-sent", sess(), cmddisp);
        }
        else
          // we have a command to send ...
          invokeEvent ("command", sess(), cmd);

        telnet->waitingForData();

        //advance to next command (that will be sent on next call)
        ++d->seqIter;

        //finish execution if there's nothing more to send
        if (d->seqIter == d->loginSeq.end())
        {
          d->connecting->stop ();
          //I have no idea why this is needed
          QObject::disconnect (d->connecting, SIGNAL (timeout()),
              this, SLOT (sendLoginAndPassword()));
          delete d->connecting;
          d->connecting = 0;
          wasData = false;
        }
      }
    }
    else
      if (telnet->newData())
      {
        telnet->waitingForData();
        wasData = true;
      }
  }
  else
  {
    d->connecting->stop ();
    //I have no idea why this is needed
    QObject::disconnect (d->connecting, SIGNAL (timeout()),
        this, SLOT (sendLoginAndPassword()));
    delete d->connecting;
    d->connecting = NULL;
    wasData = false;
  }
}

void cConnection::establishQuickConnection (cConnectionInfo *cci)
{
  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return;

  if (telnet->isConnected ())
    disconnect ();
  d->connInProgress = false;
  d->qconnInProgress = true;
  telnet->connectIt(cci->server(), cci->port ());
  d->ci = cci;

  // now wait for event from telnet
}

bool cConnection::isConnected ()
{
  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return false;

  return telnet->isConnected ();
}

void cConnection::disconnect ()
{
  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return;

  if (d->connecting != NULL)
  {
    d->connecting->stop ();
    delete d->connecting;
    d->connecting = NULL;
  }
  d->saver->stop ();
  if (d->cp != 0)
  {
    invokeEvent ("save", sess());

    // saved, now destroy all the lists
    std::map<QString, cSaveableList *>::iterator it;
    for (it = d->dataLists.begin(); it != d->dataLists.end(); ++it)
      delete it->second;
    d->dataLists.clear ();
    delete d->variablelist;
    d->variablelist = 0;
    delete d->windowlist;
    d->windowlist = 0;

    delete d->cp;
    d->cp = 0;
  }

  //clear command queue, so that pending commands don't close the tab
  //immediately
  d->commands = QString::null;

/*
//CODE DISABLED - we need connection info in case of reconnecting
  if (ci != 0)
  {
    delete ci;
    ci = 0;
  }
*/
  if (!d->destroying)
  {
    kmuddy->disableConnectedOptions ();
    kmuddy->disableAdvancedOptions ();
    kmuddy->showGauges (false);
  }
  if (telnet->isConnected())
    telnet->disconnect ();
}

void cConnection::reconnect ()
{
  if (d->ci == 0)
  {
    //We should never get here; Reconnect should ONLY be avail when it makes
    //sense; but you never know...
    KMessageBox::sorry (kmuddy, i18n ("No previous connection - cannot reconnect."), i18n ("Reconnect"));
    return;     //cannot reconnect - no connection info remembered!
  }

  //inform plug-ins
  invokeEvent ("reconnect", sess());

  //reconnect...
  if (d->ci->isQuick())
    establishQuickConnection (d->ci);
  else
    establishConnection (d->ci);
}

void cConnection::addCommand (const QString &command)
{
  QString cmd = command;
  invokeEvent ("new-command", sess(), cmd);
  if ((command != QString::null) && (cmd == QString::null)) {
    // the handler has cleared the command - do not send it
    return;
  }
  d->commands += command;
  d->commands += "\r\n";
  if (!d->waitCommands)
    sendCommands ();

  // technically, this is wrong, because the command can still be in the queue,
  // but I believe it's sufficient
  invokeEvent ("command-sent", sess(), command);
}

void cConnection::sendCommands ()
{
  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (telnet == 0)
    return;
  if (d->commands != QString::null)
    telnet->sendData (d->commands);
  d->commands = QString::null;
  d->waitCommands = false;
}

int cConnection::sentCommands ()
{
  return d->commandNum;
}

void cConnection::handleAliasesDialog ()
{
  cAliasList *al = dynamic_cast<cAliasList *>(object ("aliaslist"));
  //no aliases class => nothing to do
  if (!al)
    return;

  //so first we have to create the dialog...
  dlgAliasTrigger *adlg = new dlgAliasTrigger (al, kmuddy);

  //dialog is ready - show it!
  adlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete adlg;

  invokeEvent ("save", sess());
}

void cConnection::handleTriggersDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cTriggerList *tl = dynamic_cast<cTriggerList *>(object ("triggerlist"));
  if (!tl)
    return;

  //so first we have to create the dialog...
  dlgAliasTrigger *tdlg = new dlgAliasTrigger (tl, kmuddy);

  //dialog is ready - show it!
  tdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete tdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleActionsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cActionList *al = dynamic_cast<cActionList *>(object ("actionlist"));
  if (!al)
    return;

  //so first we have to create the dialog...
  dlgAliasTrigger *tdlg = new dlgAliasTrigger (al, kmuddy);

  //dialog is ready - show it!
  tdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete tdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleTimersDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cTimerList *tl = dynamic_cast<cTimerList *>(object ("timerlist"));
  if (!tl)
    return;

  //so first we have to create the dialog...
  dlgTimers *timdlg = new dlgTimers (tl, kmuddy);

  //dialog is ready - show it!
  timdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete timdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleShortcutsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cShortcutList *sl = dynamic_cast<cShortcutList *>(object ("shortcutlist"));
  if (!sl)
    return;

  //so first we have to create the dialog...
  dlgShortcuts *shdlg = new dlgShortcuts (sl, kmuddy);

  //dialog is ready - show it!
  shdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete shdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleVariablesDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cVariableList *vl = dynamic_cast<cVariableList *>(object ("variablelist"));
  if (!vl)
    return;

  //so first we have to create the dialog...
  dlgVariables *vdlg = new dlgVariables (vl, kmuddy);

  //dialog is ready - show it!
  vdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete vdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleVarTrigsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cVarTriggerList *vl = dynamic_cast<cVarTriggerList *>(object ("vartriggerlist"));
  if (!vl)
    return;

  //so first we have to create the dialog...
  dlgVarTrigList *vdlg = new dlgVarTrigList (vl, kmuddy);

  //dialog is ready - show it!
  vdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete vdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleStatusVarsDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cStatusVarList *svl = dynamic_cast<cStatusVarList *>(object ("statusvarlist"));
  if (!svl)
    return;

  //so first we have to create the dialog...
  dlgStatusVars *svdlg = new dlgStatusVars (svl, kmuddy);

  //dialog is ready - show it!
  svdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete svdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleGaugesDialog ()
//this is very similar to handleAliasesDialog() function above...
{
  cGaugeList *gl = dynamic_cast<cGaugeList *>(object ("gaugelist"));
  if (!gl)
    return;

  //so first we have to create the dialog...
  dlgGauges *gdlg = new dlgGauges (gl, kmuddy);

  //dialog is ready - show it!
  gdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete gdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleWindowsDialog ()
{
  cWindowList *wl = dynamic_cast<cWindowList *>(object ("windowlist"));
  if (!wl)
    return;

  dlgWindows *wdlg = new dlgWindows (wl, kmuddy);

  wdlg->exec ();

  delete wdlg;

  invokeEvent ("save", sess());
}

void cConnection::handleTickTimerDialog ()
{
  cTimerList *tl = dynamic_cast<cTimerList *>(object ("timerlist"));
  if (!tl) return;
  cTimerTick *tt = tl->getTimerTick();
  if (!tt) return;

  //so first we have to create the dialog...
  dlgTimerTick *ttdlg = new dlgTimerTick (tt, kmuddy);

  //dialog is ready - show it!
  ttdlg->exec ();

  //further action is handled by the dialog itself, so that we only
  //have to destroy the dialog...
  delete ttdlg;

  invokeEvent ("save", sess());
}

void cConnection::switchAliases (bool val)
{
  cAliasList *al = dynamic_cast<cAliasList *>(object ("aliaslist"));
  if (!al)
    return;
  al->setEnabled (val);
  invokeEvent ("message", sess(),
      val ? i18n ("Aliases are now enabled.") : i18n ("Aliases are now disabled."));
}

void cConnection::switchTriggers (bool val)
{
  cTriggerList *tl = dynamic_cast<cTriggerList *>(object ("triggerlist"));
  if (!tl)
    return;
  tl->setEnabled (val);
  invokeEvent ("message", sess(),
      val ? i18n ("Triggers are now enabled.") : i18n ("Triggers are now disabled."));
}

void cConnection::switchTimers (bool val)
{
  cTimerList *tl = dynamic_cast<cTimerList *>(object ("timerlist"));
  if (!tl)
    return;
  if (val)
    tl->enableTimers ();
  else
    tl->disableTimers ();
  invokeEvent ("message", sess(),
      val ? i18n ("Timers are now enabled.") : i18n ("Timers are now disabled."));
}

void cConnection::switchShortcuts (bool val)
{
  cShortcutList *sl = dynamic_cast<cShortcutList *>(object ("shortcutlist"));
  if (!sl)
    return;
  sl->setEnabled (val);
  invokeEvent ("message", sess(),
      val ? i18n ("Macro keys are now enabled.") : i18n ("Macro keys are now disabled."));
}

void cConnection::updateMenus ()
{
  // do nothing if this session is not active
  if (sess() != cSessionManager::self()->activeSession())
    return;

  // TODO: reimplent using the modern architecture
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet", sess()));
  if (!telnet)
    return;
  if (telnet->isConnected())
  {
    kmuddy->unsetClosedConn ();
    kmuddy->enableConnectedOptions();
    if (d->cp)
      kmuddy->enableAdvancedOptions();
    else
      kmuddy->disableAdvancedOptions();

    if (d->cp)
    {
      cSaveableList *al = getList ("aliases");
      cSaveableList *tl = getList ("triggers");
      cSaveableList *ml = getList ("timers");
      cSaveableList *sl = getList ("shortcuts");
      if (al) kmuddy->setMenuAliasesEnabled (al->enabled ());
      if (tl) kmuddy->setMenuTriggersEnabled (tl->enabled ());
      if (ml) kmuddy->setMenuTimersEnabled (ml->enabled ());
      if (sl) kmuddy->setMenuShortcutsEnabled (sl->enabled ());
    }
  }
  else
  {
    kmuddy->disableConnectedOptions();
    if (d->connclosed)
      kmuddy->setClosedConn (d->canCT);
    else
      kmuddy->unsetClosedConn ();
  }
}

void cConnection::loadLists ()
{
  if (!d->cp)
    return;

  QString directory = locateLocal ("appdata",
      "profiles/" + d->cp->name() + "/");
  QString fname = directory + "aliases";
  cAliasList *aliaslist = new cAliasList (fname, sess());
  d->dataLists["aliases"] = aliaslist;
  fname = directory + "triggers";
  cTriggerList *triggerlist = new cTriggerList (fname, sess());
  d->dataLists["triggers"] = triggerlist;
  fname = directory + "actions";
  cActionList *actionlist = new cActionList (fname, sess());
  d->dataLists["actions"] = actionlist;
  actionlist->rebuildToolbar ();
  fname = directory + "aliasgroups";
  d->dataLists["aliasgroups"] = new cGroupList (fname, "aliasgrouplist", sess(), aliaslist);
  fname = directory + "triggergroups";
  d->dataLists["triggergroups"] = new cGroupList (fname, "triggergrouplist", sess(), triggerlist);
  fname = directory + "timers";
  cTimerList *timerlist = new cTimerList (fname, sess());
  d->dataLists["timers"] = timerlist;
  timerlist->enableTimers ();
  fname = directory + "macrokeys";
  d->dataLists["shortcuts"] = new cShortcutList (fname, sess());
  fname = directory + "variables";
  d->variablelist = new cVariableList (fname, sess());
  fname = directory + "vartriggers";
  d->dataLists["vartrigs"] = new cVarTriggerList (fname, sess());
  fname = directory + "statusvars";
  d->dataLists["statusvars"] = new cStatusVarList (fname, sess());
  fname = directory + "gauges";
  d->dataLists["gauges"] = new cGaugeList (fname, sess());
  fname = directory + "windowlist";
  d->windowlist = new cWindowList (fname, sess());
}

void cConnection::saverTimeout ()
{
  invokeEvent ("save", sess());
}

void cConnection::save ()
{
  if (d->cp == 0) return;
  d->cp->save ();
  //save all the lists
  std::map<QString, cSaveableList *>::iterator it;
  for (it = d->dataLists.begin(); it != d->dataLists.end(); ++it)
    it->second->save ();
  d->variablelist->save();
  d->windowlist->save();
}

