//
// C++ Implementation: cSessionManager
//
// Description: Session manager.
//
//
// Author: Tomas Mecir <kmuddy@kmuddy.net>, (C) 2002-2006
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "csessionmanager.h"

#include "cconnection.h"
#include "cglobalsettings.h"
#include "cinputline.h"
#include "csession.h"
#include "ctabbar.h"
#include "ctelnet.h"
#include "kmuddy.h"

#include <qiconset.h>
#include <qwidgetstack.h>
#include <kapp.h>
#include <kiconloader.h>
#include <klocale.h>

extern KMuddy *kmuddy;
extern cGlobalSettings *globalsettings;

struct cSessionManagerPrivate {
  int count;
  int tabID[MAXSESSIONS];
  int wintabID[MAXSESSIONS];
  
  cActionManager *am;
  QWidgetStack *stack;
  cTabBar *tabbar;

  /** icons for the tab bar */
  QIconSet iconOk, iconNo, iconFlag;

  /** should the tab bar be displayed even if there is only one connection? */
  bool wantTabBar;
};

cSessionManager *cSessionManager::_self = 0;

cSessionManager::cSessionManager ()
{
  d = new cSessionManagerPrivate;
  d->count = 0;
  cActionManager::self()->setActiveSession (0);
  d->am = cActionManager::self();

  //icons for tabs
  KIconLoader *loader = (KApplication::kApplication())->iconLoader();
  QPixmap i1 = loader->loadIcon ("ok", KIcon::Small, 16);
  QPixmap i2 = loader->loadIcon ("cancel", KIcon::Small, 16);
  QPixmap i3 = loader->loadIcon ("flag", KIcon::Small, 16);
  d->iconOk.setPixmap (i1, QIconSet::Automatic);
  d->iconNo.setPixmap (i2, QIconSet::Automatic);
  d->iconFlag.setPixmap (i3, QIconSet::Automatic);
}

cSessionManager::~cSessionManager ()
{
  delete d;
}

cSessionManager *cSessionManager::self()
{
  if (!_self) _self = new cSessionManager;
  return _self;
}

int cSessionManager::count ()
{
  return d->count;
}

int cSessionManager::activeSession ()
{
  return cActionManager::self()->activeSession();
}

int cSessionManager::getSessionByTab (int tabnumber)
{
  //find session ID; a bit ineffective, but MAXSESSIONS is only 50,
  //so nobody will even notice the slowdown...
  int which = -1;
  for (int i = 1; i <= MAXSESSIONS; i++)
    if (d->tabID[i] == tabnumber)
  {
    which = i;
    break;
  }
  return which;
}

int cSessionManager::addSession (bool profile)
{
  if (KMuddy::isGoingDown())
    return 0;

  if (d->count == MAXSESSIONS)
    return 0;

  //first, one very special case - if we only have one session
  //and that one is not connected, we return that session
  //and don't add any new one. This session will be used for
  //the connection that is being established.
  if (d->count == 1)
  {
    int which = 0;
    for (int i = 1; i <= MAXSESSIONS; i++)
      if (d->am->sessionExists (i))
      {
        which = i;
        break;
      }
    cTelnet *telnet = dynamic_cast<cTelnet *>(d->am->object ("telnet", which));
    if (!telnet->isConnected())
    {
      cActionManager::self()->setSessionAttrib (which, "profile", profile?1:0);
      globalsettings->apply ();
      return which;
    }
    //else: continue
  }

  //now for the standard operation

  //find a first free place for the connection
  int which = -1;
  for (int i = 1; i <= MAXSESSIONS; i++)
    if (!d->am->sessionExists (i))
  {
    which = i;
    break;
  }
  if (which == -1)  //this means a BUG in program, hope it'll never happen
    return 0;

  d->am->registerSession (which);
  cActionManager::self()->setSessionAttrib (which, "profile", profile?1:0);
  cSession *session = new cSession (which, d->stack);
  d->stack->addWidget (session, which);
  d->count++;

  // inform everyone
  d->am->invokeEvent ("session-created", 0, which);
  d->am->invokeEvent ("created", which);
  
  //update tab bar
  QTab *tab = new QTab (d->iconNo, i18n ("No connection"));
  d->tabID[which] = d->tabbar->addTab (tab);
  d->tabbar->update ();

  setSession (which);
  //just to be sure...
  cActionManager::self()->setActiveSession (which);

  (KApplication::kApplication())->processEvents();

  //apply global settings to the newly created session
  //this also shows/hides the tabbar if needed
  globalsettings->apply ();

  return which;
}

bool cSessionManager::removeSession (int which, bool dontClose)
{
  if (KMuddy::isGoingDown())
    return false;

  //reconnect shortcut
  QString reconnectText = kmuddy->reconnectText ();
  
  if ((which <= 0) || (which > MAXSESSIONS))
    return false;
  if (!d->am->sessionExists (which))
    return false;

  //now the special case - only one session; we do almost nothing!
  if (d->count == 1)
  {
    setSessionName (which, i18n ("No connection"));
    //no connection...
    d->tabbar->tab (d->tabID[which])->setIconSet (d->iconNo);
    d->tabbar->update ();
    cConnection *connection = dynamic_cast<cConnection *>(d->am->object ("connection", which));
    if (connection->getConnectionInfo())
    {
      connection->setConnectionClosed (false);
      connection->updateMenus ();
      d->am->invokeEvent ("message", which, reconnectText);
    }
    if (!d->wantTabBar)
      d->tabbar->hide ();
    return true;
  }

  if (!dontClose)
  {
    //inform everyone
    d->am->invokeEvent ("session-destroyed", 0, which);
    d->am->invokeEvent ("destroyed", which);
    //remove widget
    d->stack->removeWidget (d->stack->widget (which));
    cSession *sess = dynamic_cast<cSession *>(d->am->object ("session", which));
    delete sess;
    d->am->unregisterSession (which);
    d->count--;

    //update tab bar
    d->tabbar->removeTab (d->tabbar->tab (d->tabID[which]));

    //removeTab switches to some other tab but doesn't emit a signal (no idea why)
    //so I have to do this myself :(((
    int id = d->tabbar->currentTab ();
    changeSession (id);

    //hide the tabbar if needed
    if ((!d->wantTabBar) && (d->count <= 1))
      d->tabbar->hide ();
  }
  else
  {
    setSessionName (which, i18n ("Connection closed."));
    d->tabbar->tab (d->tabID[which])->setIconSet (d->iconNo);
    d->tabbar->update ();
    cConnection *connection = dynamic_cast<cConnection *>(d->am->object ("connection", which));
    connection->setConnectionClosed (true);
    connection->updateMenus ();
    d->am->invokeEvent ("message", which, reconnectText);
  }

  return true;
}

void cSessionManager::setSession (int which)
{
  if ((which <= 0) || (which > MAXSESSIONS))
    return;
  if (d->am->sessionExists (which))
    return;
  //this also invokes changeSession, which updates active session
  d->tabbar->setCurrentTab (d->tabID[which]);

  //do not remove this line! It's usually useless, but it's required
  //for the program to work correctly on startup with hidden tabbar
  d->stack->raiseWidget (which);
}

void cSessionManager::changeSession (int id)
{
  int which = getSessionByTab (id);
  if (which == -1)  //such session does not exist
    return;

  d->am->invokeEvent ("deactivated", which);
  d->am->invokeEvent ("session-deactivated", 0, which);

  d->stack->raiseWidget (which);
  cActionManager::self()->setActiveSession (which);
  QTab *tab = d->tabbar->tab (d->tabID[which]);

  //disable flashing (if enabled)
  cSession *sess = dynamic_cast<cSession *>(d->am->object ("session", which));
  if (sess && sess->flashing())
  {
    sess->setFlashing (false);

    //if flashing was on, icon could be wrong - fix it
    cTelnet *telnet = dynamic_cast<cTelnet *>(d->am->object ("telnet", which));
    if (!telnet->isConnected ())
      tab->setIconSet (d->iconNo);
    else
      tab->setIconSet (d->iconOk);
    d->tabbar->update();
  }

  d->am->invokeEvent ("session-activated", 0, which);
  d->am->invokeEvent ("activated", which);

  //global caption
  kmuddy->setPlainCaption (kapp->makeStdCaption (tab->text ()));

  //update menus (active items and such)
  int s = activeSession();
  cConnection *connection = dynamic_cast<cConnection *>(d->am->object ("connection", s));
  connection->updateMenus ();

  // update windows
  kmuddy->updateWindows ();

  //give focus to the input line
  //cInputLine::focus[In/Out]Event also restores selections :)
  cInputLine *inputline = dynamic_cast<cInputLine *>(d->am->object ("inputline", s));
  inputline->setFocus ();
}

void cSessionManager::setSessionName (int which, QString name, bool defName)
{
  cSession *sess = dynamic_cast<cSession *>(d->am->object ("session", which));
  sess->setName (name, defName);
  //update the name - maybe it wasn't changed?
  name = sess->name();
  QTab *tab = d->tabbar->tab (d->tabID[which]);
  tab->setText (name);
  //change icon to Ok (we'll change it later if it's not correct)
  tab->setIconSet (d->iconOk);
  d->tabbar->update ();
  if (which == activeSession ())
    kmuddy->setPlainCaption (kapp->makeStdCaption (name));
}

void cSessionManager::setWidgetStack (QWidgetStack *stack)
{
  d->stack = stack;
}

void cSessionManager::setTabBar (cTabBar *tabbar)
{
  d->tabbar = tabbar;
}

bool cSessionManager::alwaysTabBar ()
{
  return d->wantTabBar;
}

void cSessionManager::setAlwaysTabBar (bool value)
{
  if (d->tabbar == 0)
    return;
  d->wantTabBar = value;
  if (value)
    d->tabbar->show ();
  else
    if (d->count > 1)
      d->tabbar->show ();
  else
    d->tabbar->hide ();
}

void cSessionManager::setNotifyFlag (int sess)
{
  cSession *session = dynamic_cast<cSession *>(d->am->object ("session", sess));
  session->setFlashing (true);

  QTab *tab = d->tabbar->tab (d->tabID[sess]);
  cConnection *connection = dynamic_cast<cConnection *>(d->am->object ("connection", sess));
  bool connClosed = connection->connectionClosed();
  tab->setIconSet(connClosed ? d->iconNo : d->iconFlag);
  d->tabbar->update();
}


