/***************************************************************************
                          ctelnet.cpp  -  telnet...
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Pi Jun 14 2002
    copyright            : (C) 2002-2005 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#define CTELNET_CPP

#include "ctelnet.h"

#include "cactionmanager.h"
#include "cconnprefs.h"
#include "cmccp.h"

#ifdef HAVE_SOUND
#include "cmsp.h"
#endif

// needed for removeSession which is called in disconnect
#include "csessionmanager.h"

#ifdef HAVE_MXP
#include "cmxpmanager.h"
#endif

#include <klocale.h>
#include <kbufferedsocket.h>
#include <kmessagebox.h>
#include <kapplication.h>

#include <stdio.h>

struct cTelnetPrivate {
  /** socket */
  KNetwork::KBufferedSocket *socket;
  QString hostName;
  int hostPort;
  
  /** object that handles MCCP */
  cMCCP *MCCP;

#ifdef HAVE_SOUND
  /** object that handles MSP */
  cMSP *MSP;
  bool usingmsp;
#endif

#ifdef HAVE_MXP
  bool usingmxp, mxpNegotiated;
  int mxpallow;
#endif
  
  //iac: last char was IAC
  //iac2: last char was DO, DONT, WILL or WONT
  //insb: we're in IAC SB, waiting for IAC SE
  QString command;
  bool iac, iac2, insb;
  
  /** has any data arrived since last call to waitingForData() */
  bool newdata;
  /** are we connected or commecting? */
  bool _connected, _connecting;
  
  /** current state of options on our side and on server side */
  bool myOptionState[256], hisOptionState[256];
    /** whether we have announced WILL/WON'T for that option (if we have, we don't
  respond to DO/DON'T sent by the server -- see implementation and RFC 854
  for more information... */
  bool announcedState[256];
  /** whether the server has already announced his WILL/WON'T */
  bool heAnnouncedState[256];
  /** whether we have tried to enable this option */
  bool triedToEnable[256];
  /** amount of bytes sent up to now */
  int sentbytes;
  /** have we received the GA signal? */
  bool recvdGA;
  /** should we prepend newline after receving a GA */
  bool prependGANewLine;
  bool t_cmdEcho;
  bool t_lpmudstyle;
  bool _startupneg;
  /** current dimensions */
  int curX, curY;
  /** offline connection */
  bool offLineConnection;

  QString quitcommand, termType;
};

cTelnet::cTelnet (int sess) : cActionBase ("telnet", sess)
{
  d = new cTelnetPrivate;

  d->socket = 0;
  d->termType = "KMuddy";
  
  d->iac = d->iac2 = d->insb = false;
  d->command = "";

  d->sentbytes = 0;
  d->offLineConnection=false;
  d->_connected = false;
  d->_connecting = false;
  d->curX = 0;
  d->curY = 0;
  d->_startupneg = true;
  setQuitCommand (QUIT);
  
  // initialize the socket
  createSocket ();

  reset ();

  d->MCCP = new cMCCP (this);

#ifdef HAVE_SOUND
  d->MSP = new cMSP (sess);
  d->usingmsp = false;
#endif

#ifdef HAVE_MXP
  d->usingmxp = false;
  d->mxpallow = 3;  //auto-detect
  d->mxpNegotiated = false;
#endif

  addEventHandler ("dimensions-changed", 20, PT_INT);
}

cTelnet::~cTelnet()
{
  if (isConnected ())
    disconnect ();

  removeEventHandler ("dimensions-changed");

  delete d->socket;
  delete d->MCCP;

#ifdef HAVE_SOUND
  delete d->MSP;
#endif

  delete d;
  d = 0;
}

void cTelnet::eventHandler (QString event, int, int par1, int par2)
{
  if (event == "dimensions-changed") {
    windowSizeChanged (par1, par2);
  }
}

QString cTelnet::actionHandler (QString action, int,
    const QString &par1, const QString &)
{
  if (action == "set-quit-command")
    setQuitCommand (par1);
  return QString::null;
}

QString cTelnet::actionHandler (QString action, int, int par1, int)
{
  if (action == "mxp-allowed") {
#ifdef HAVE_MXP
    setMXPAllowed (par1);
#endif
  }
  else if (action == "set-negotiate-on-startup") {
    setNegotiateOnStartup (par1 != 0);
  }
  else if (action == "set-lp-mud-style") {
    setLPMudStyle (par1 != 0);
  }
  return QString::null;
}

void cTelnet::connectionFailed ()
{
#if KDE_IS_VERSION(3,4,0)
  if (d->socket->error() == KNetwork::KBufferedSocket::RemotelyDisconnected)
  {
    while (d->socket->bytesAvailable())
      socketRead ();
    d->socket->closeNow ();
    return;
  }
#endif
  
  // don't crash on weird problems (happens on KDE 3.4, no idea why)
  if (d->socket->error() == KNetwork::KBufferedSocket::NoError)
    return;
 
  QString err = d->socket->errorString();
  cActionManager::self()->invokeEvent ("connection-failed", sess(), err);
  d->_connected = false;
  d->_connecting = false;
}

/** establishes a new connection */
void cTelnet::connectIt (const QString &address, int port)
{
  // close existing connection first (if any)
  if (isConnected ())
    disconnect ();

  // handle offline connection
  if (isOffLineConnection())
  {
    d->_connected = true;
    d->_connecting = false;
    invokeEvent ("message", sess(), i18n ("--- A connection has been established ---"));
    invokeEvent ("connected", sess());
    return;
  }

  d->_connecting = true;
  cActionManager::self()->invokeEvent ("message", sess(), i18n ("Connecting ..."));
  d->socket->setTimeout (10000);
  d->hostName = address;
  d->hostPort = port;
  d->socket->connect (address, QString::number (port));
}

void cTelnet::socketFailed (int)
{
  connectionFailed ();
}

void cTelnet::createSocket ()
{
  d->socket = new KNetwork::KBufferedSocket;
  connect (d->socket, SIGNAL(connected(const KResolverEntry &)), this, SLOT (socketConnected ()));
  connect (d->socket, SIGNAL(gotError(int)), this, SLOT (socketFailed (int)));
  connect (d->socket, SIGNAL(readyRead()), this, SLOT (socketRead ()));
  connect (d->socket, SIGNAL(closed()), this, SLOT (socketClosed ()));
  d->socket->enableRead (true);
  d->socket->setInputBuffering (true);
  d->socket->setOutputBuffering (true);
}

void cTelnet::reset ()
{
  puts ("KMuddy: Telnet reset.");
  //prepare option variables
  for (int i = 0; i < 256; i++)
  {
    d->myOptionState[i] = false;
    d->hisOptionState[i] = false;
    d->announcedState[i] = false;
    d->heAnnouncedState[i] = false;
    d->triedToEnable[i] = false;
  }
  //reset telnet status
  d->iac = d->iac2 = d->insb = false;
  d->command = "";
}

void cTelnet::socketConnected ()
{
  if (d->_connected) return;

  d->_connected = true;
  d->_connecting = false;

  //reset MCCP, MSP and byte counters
  d->MCCP->reset ();
#ifdef HAVE_SOUND
  d->MSP->reset (d->hostName);
#endif
  
  reset ();

  //now we should be connected
  invokeEvent ("message", sess(), i18n ("--- A connection has been established ---"));
  cActionManager::self()->invokeEvent ("connected", sess());

#ifdef HAVE_MXP
  // MXP -must- be initialized AFTER the connected message, else it won't work
  setMXPAllowed (d->mxpallow);  //initialize MXP properly
#endif  
  d->sentbytes = 0;
  
  //negotiate some telnet options, if allowed
  if (d->_startupneg)
  {
    QString s;
    
    //NAWS (used to send info about window size)
    s += (QChar) TN_IAC;
    s += (QChar) TN_WILL;
    s += (QChar) OPT_NAWS;
    //do not allow server to echo our text!
    s += (QChar) TN_IAC;
    s += (QChar) TN_DONT;
    s += (QChar) OPT_ECHO;
  
    doSendData (s, false);
  }
}

/** closes connection */
void cTelnet::disconnect ()
{
  if (!d->_connected) return;

  // wait a bit to see if we get any more data, process it if we do ...
  d->_connected = false;
  d->_connecting = false;

  reset ();

  // send command quit first! - otherwise MUD would say you're net-dead...
  if (isConnected () && !isOffLineConnection())
    doSendData (d->quitcommand + "\n", true);

  if (isOffLineConnection())
    d->offLineConnection=false;

  //reset quit-command back to default value
  d->quitcommand = QUIT;

  d->socket->flush ();
  d->socket->closeNow ();

  //alright - we're disconnected
  cActionManager::self()->invokeEvent ("disconnected", sess());

  // ------------------
  // TODO: the rest will not be here, but in their respective objects, hooked on the event
  
  //remove the session - should now be safe
  cSessionManager::self()->removeSession (sess(), true);
}

void cTelnet::setOffLineConnection (bool type)
{
  d->offLineConnection = type;
}

/** are we connected (offline connection is counted as connected) ? */
bool cTelnet::isConnected ()
{
  return (isOffLineConnection() || d->_connected);
}

#ifdef HAVE_SOUND

void cTelnet::setMSPGlobalPaths (const QStringList &paths)
{
  if (d->MSP != 0)
    d->MSP->setGlobalPaths (paths);
}

bool cTelnet::usingMSP ()
{
  return d->usingmsp;
}

void cTelnet::setMSPAllowed (bool allow)
{
  if (d->MSP != 0)
    d->MSP->setMSPAllowed (allow);
}

void cTelnet::setDownloadAllowed (bool allow)
{
  if (d->MSP != 0)
    d->MSP->setDownloadAllowed (allow);
}

void cTelnet::processSoundRequest (bool isSOUND, QString fName, int volume, int repeats,
    int priority, QString type, QString url)
{
  if (d->MSP != 0)
    d->MSP->processRequest (isSOUND, fName, volume, repeats, priority, type, url);
}

#endif  //HAVE_SOUND

#ifdef HAVE_MXP

bool cTelnet::usingMXP () {
  return d->usingmxp;
}

int cTelnet::MXPAllowed () {
  return d->mxpallow;
}

void cTelnet::setMXPAllowed (int allow)
{
  //update actual MXP usage...
  switch (allow) {
    case 1: //Never
    {
      callAction ("mxpmanager", "set-active", sess(), 0);
      //ask server to disable MXP
      QString s = (QChar) TN_IAC;
      s += (QChar) TN_DONT;
      s += (QChar) OPT_MXP;
      doSendData (s, false);
    }
    break;
    case 2: //If negotiated
    {
      callAction ("mxpmanager", "set-active", sess(), d->mxpNegotiated ? 1 : 0);
      callAction ("mxpmanager", "switch-open", sess());
    }
    break;
    case 3: //Auto-detect
    {
      callAction ("mxpmanager", "set-active", sess(), 1);
    }
    break;
    case 4: //Always
    {
      callAction ("mxpmanager", "set-active", sess(), 1);
      callAction ("mxpmanager", "switch-open", sess());
    }
    break;
  };
  d->mxpallow = allow;
}

#endif //HAVE_MXP

bool cTelnet::sendData (const QString &data)
{
  if (!(isConnected()))
    return false;
  // return true and dont send, since offline is used
  if (isOffLineConnection())
    return true;

  if (d->t_cmdEcho == true && d->t_lpmudstyle)
    d->prependGANewLine = false;

  //IAC byte must be doubled
  QString d;
  int len = data.length();
  bool gotIAC = false;
  for (int i = 0; i < len; i++)
    if ((unsigned char) data[i].latin1() == 255) {
      gotIAC = true;
      break;
    }
  if (gotIAC)  // double IACs, if any
    for (int i = 0; i < len; i++)
    {
      QChar ch = data[i];
      d += ch;
      if ((unsigned char) ch.latin1() == 255)
        d += ch;  //double IAC
    }

  //data ready, send it!
  // don't use 8-bit if we had to expand IACs
  return doSendData (gotIAC ? d : data, gotIAC ? false : true);
}

void cTelnet::waitingForData () {
  d->newdata = false;
}

bool cTelnet::newData () {
  return d->newdata;
}

bool cTelnet::doSendData (const QString &data, bool use8Bit)
{
  if (!(isConnected()))
    return false;
  if (isOffLineConnection())
    return true;
  QString dd = data;
  //write data to socket - it's so complicated because sometimes only a part of data
  //is accepted at a time
  int dataLength = data.length ();
  int remlen = dataLength;
  do {
    int written;
    if (use8Bit)
      written = d->socket->writeBlock (dd.local8Bit (), remlen);
    else
      written = d->socket->writeBlock (dd.latin1 (), remlen);
    if (written == -1)  // buffer full - try again
      continue;
    remlen -= written;
    dd.remove (0, written);
  } while (remlen > 0);

  //update counter
  d->sentbytes += dataLength;
  return true;
}

void cTelnet::windowSizeChanged (int x, int y)
{
  //remember the size - we'll need it if NAWS is currently disabled but will
  //be enabled. Also remember it if no connection exists at the moment;
  //we won't be called again when connecting
  d->curX = x;
  d->curY = y;
  if (!(isConnected()))
    return;
  if (d->myOptionState[OPT_NAWS])   //only if we have negotiated this option
  {
    QString s;
    s = (QChar) TN_IAC;
    s += (QChar) TN_SB;
    s += (QChar) OPT_NAWS;
    unsigned char x1, x2, y1, y2;
    x1 = (unsigned char) x / 256;
    x2 = (unsigned char) x % 256;
    y1 = (unsigned char) y / 256;
    y2 = (unsigned char) y % 256;
    //IAC must be doubled
    s += (QChar) x1;
    if (x1 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) x2; 
    if (x2 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) y1;
    if (y1 == TN_IAC)
      s += (QChar) TN_IAC;
    s += (QChar) y2;
    if (y2 == TN_IAC)
      s += (QChar) TN_IAC;
    
    s += (QChar) TN_IAC;
    s += (QChar) TN_SE;
    doSendData (s, false);
  }
}

void cTelnet::sendTelnetOption (int type, int option)
{
  QString s;
  s = (QChar) TN_IAC;
  s += (QChar) type;
  s += (QChar) option;
  doSendData (s, false);
}

void cTelnet::processTelnetCommand (const QString &command)
{
  unsigned char ch = command[1].latin1();
    unsigned char option;
    switch (ch) {
    case TN_AYT:
      doSendData ("I'm here! Please be more patient!\r\n");
          //well, this should never be executed, as the response would probably
          //be treated as a command. But that's server's problem, not ours...
          //If the server wasn't capable of handling this, it wouldn't have
          //sent us the AYT command, would it? Impatient server = bad server.
          //Let it suffer! ;-)
      break;
    case TN_GA:
      d->recvdGA = true;
      //signal will be emitted later
      break;
    case TN_WILL:
      //server wants to enable some option (or he sends a timing-mark)...
      option = command[2].latin1();

      d->heAnnouncedState[option] = true;
      if (d->triedToEnable[option])
      {
        d->hisOptionState[option] = true;
        d->triedToEnable[option] = false;
      }
      else
      {
        if (!d->hisOptionState[option])
            //only if this is not set; if it's set, something's wrong wth the server
            //(according to telnet specification, option announcement may not be
            //unless explicitly requested)
        {
          if ((option == OPT_SUPPRESS_GA) || (option == OPT_STATUS) ||
              (option == OPT_TERMINAL_TYPE) || (option == OPT_NAWS))
                 //these options are supported; compression is handled
                 //separately
          {
            sendTelnetOption (TN_DO, option);
            d->hisOptionState[option] = true;
          }
#ifdef HAVE_MXP
          else
          if (option == OPT_MXP)
          {
            // allow or disallow, MXP depending on whether it's disabled
            sendTelnetOption ((d->mxpallow >= 2) ? TN_DO : TN_DONT, option);
            d->hisOptionState[option] = true;

            //MXP is now negotiated
            d->mxpNegotiated = true;
            if (d->mxpallow == 2)  //MXP: if negotiated
            {
              cMXPManager *mm = dynamic_cast<cMXPManager *>(object ("mxpmanager"));
              mm->setMXPActive (true);
              puts ("KMuddy: MXP enabled !");
            }
          }
#endif
#ifdef HAVE_SOUND
          else
          if (option == OPT_MSP)
          {
            sendTelnetOption (TN_DO, option);
            d->hisOptionState[option] = true;

            //MSP is now enabled
            d->usingmsp = true;
            d->MSP->enableMSP ();
            puts ("KMuddy: MSP enabled !");
          }
#endif  //HAVE_SOUND
          else
          if ((option == OPT_COMPRESS) || (option == OPT_COMPRESS2))
          //these are handled separately, as they're a bit special
          {
            if ((option == OPT_COMPRESS) && (d->hisOptionState[OPT_COMPRESS2]))
            {
              //protocol says: reject MCCP v1 if you have previously accepted
              //MCCP v2...
              sendTelnetOption (TN_DONT, option);
              d->hisOptionState[option] = false;
              puts ("KMuddy: Rejecting MCCP v1, because v2 is already used !");
            }
            else
            {
              sendTelnetOption (TN_DO, option);
              d->hisOptionState[option] = true;
              //inform MCCP object about the change
              if ((option == OPT_COMPRESS)) {
                d->MCCP->setMCCP1 (true);
                puts ("KMuddy: MCCP v1 enabled !");
              }
              else {
                d->MCCP->setMCCP2 (true);
                puts ("KMuddy: MCCP v2 enabled !");
              }
            }
          }
          else
          {
            sendTelnetOption (TN_DONT, option);
            d->hisOptionState[option] = false;
          }
        }
      }
      break;
    case TN_WONT:
      //server refuses to enable some option...
      option = command[2].latin1();
      if (d->triedToEnable[option])
      {
        d->hisOptionState[option] = false;
        d->triedToEnable[option] = false;
        d->heAnnouncedState[option] = true;
      }
      else
      {
        //send DONT if needed (see RFC 854 for details)
        if (d->hisOptionState[option] || (!d->heAnnouncedState[option]))
        {
          sendTelnetOption (TN_DONT, option);
          d->hisOptionState[option] = false;
          //inform MCCP object about the change - won't cause problems in
          //cMCCP - see cmccp.cpp for more info
          if ((option == OPT_COMPRESS)) {
            d->MCCP->setMCCP1 (false);
            puts ("KMuddy: MCCP v1 disabled !");
          }
          if ((option == OPT_COMPRESS2)) {
            d->MCCP->setMCCP2 (false);
            puts ("KMuddy: MCCP v1 disabled !");
          }
        }
        d->heAnnouncedState[option] = true;
      }
#ifdef HAVE_SOUND
      if (option == OPT_MSP)
      {
        //MSP is now disabled
        d->usingmsp = false;
        d->MSP->disableMSP ();
        puts ("KMuddy: MSP disabled !");
      }
#endif
#ifdef HAVE_MXP
      if (option == OPT_MXP)
      {
        //MXP is now disabled
        cMXPManager *mm = dynamic_cast<cMXPManager *>(object ("mxpmanager"));
        mm->setMXPActive (false);
        puts ("KMuddy: MXP disabled !");
      }
#endif
      break;
    case TN_DO:
      //server wants us to enable some option
      option = command[2].latin1();
      if (option == OPT_TIMING_MARK)
      {
        //send WILL TIMING_MARK
        sendTelnetOption (TN_WILL, option);
      }
      else if (!d->myOptionState[256])
      //only if the option is currently disabled
      {
        if ((option == OPT_SUPPRESS_GA) || (option == OPT_STATUS) ||
            (option == OPT_TERMINAL_TYPE) || (option == OPT_NAWS))
        {
          sendTelnetOption (TN_WILL, option);
          d->myOptionState[option] = true;
          d->announcedState[option] = true;
        }
        else
        {
          sendTelnetOption (TN_WONT, option);
          d->myOptionState[option] = false;
          d->announcedState[option] = true;
        }
      }
      if (option == OPT_NAWS)  //NAWS here - window size info must be sent
        windowSizeChanged (d->curX, d->curY);
      break;
    case TN_DONT:
      //only respond if value changed or if this option has not been announced yet
      option = command[2].latin1();
      if (d->myOptionState[option] || (!d->announcedState[option]))
      {
        sendTelnetOption (TN_WONT, option);
        d->announcedState[option] = true;
      }
      d->myOptionState[option] = false;
      break;
    case TN_SB:
      //subcommand - we analyze and respond...
      option = command[2].latin1();
      switch (option) {
        case OPT_STATUS:
          //see OPT_TERMINAL_TYPE for explanation why I'm doing this
          if (true /*myOptionState[OPT_STATUS]*/)
          {
            if (command[3].latin1() == TNSB_SEND)
            //request to send all enabled commands; if server sends his
            //own list of commands, we just ignore it (well, he shouldn't
            //send anything, as we do not request anything, but there are
            //so many servers out there, that you can never be sure...)
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_SB;
              s += (QChar) OPT_STATUS;
              s += (QChar) TNSB_IS;
              for (int i = 0; i < 256; i++)
              {
                if (d->myOptionState[i])
                {
                  s += (QChar) TN_WILL;
                      //nice line, isn't it? :)
                  s += (QChar) ((unsigned char) i);
                }
                if (d->hisOptionState[i])
                {
                  s += (QChar) TN_DO;
                  s += (QChar) ((unsigned char) i);
                }
              }
              s += (QChar) TN_IAC;
              s += (QChar) TN_SE;
              doSendData (s, false);
            }
          }
        break;
        case OPT_TERMINAL_TYPE:
          if (d->myOptionState[OPT_TERMINAL_TYPE])
          {
            if (command[3].latin1() == TNSB_SEND)
              //server wants us to send terminal type; he can send his own type
              //too, but we just ignore it, as we have no use for it...
            {
              QString s;
              s = (QChar) TN_IAC;
              s += (QChar) TN_SB;
              s += (QChar) OPT_TERMINAL_TYPE;
              s += (QChar) TNSB_IS;
              s += d->termType;
              s += (QChar) TN_IAC;
              s += (QChar) TN_SE;
              doSendData (s, false);
            }
          }
        break;
        //other cmds should not arrive, as they were not negotiated.
        //if they do, they are merely ignored
      };
      break;
    //other commands are simply ignored (NOP and such, see .h file for list)
  };
}

void cTelnet::socketRead ()
{
  /*
  This function makes heavy use of char* instead of relying on QString,
  because otherwise there are problems with locale-based data and/or
  telnet commands.
  */

  // Looks like we often get this before getting the connected event ...
  if (d->_connecting && (!d->_connected))
    socketConnected();

  char buffer[32769];  //we should never receive such a big amount of data,
        //but we want to be sure that we read all we can...

  char data[32769]; //clean data after decompression

  int amount = d->socket->readBlock (buffer, 32768);
  if (amount == -1)
    return;   //something is wrong (no data?)
  if (amount == 0)
    return;   //0 means socket has been closed; maybe I'll do something more?
  buffer[amount] = '\0';    //just to be sure - mark end of string

  invokeEvent ("raw-data-comp", sess(), QString (buffer));

  //we'll need cConnPrefs later on
  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));

  //MCCP will be handled first. This is done in a separate class, which
  //detects all MCCP-related codes itself. I know that this involves some
  //code duplication (two telnet option parsers), but it's needed because
  //telnet sequences may (and will) appear inside the compressed streams,
  //so it's best to get plain uncompressed data first. Another reason is
  //improper design of MCCP v1, which uses unterminated telnet subsequences.
  //This has been fixed in MCCP v2, but some MUDs may still use the old
  //version.

  d->MCCP->prepareDecompression (buffer, data, amount, 32768);
  int datalen;
  while ((datalen = d->MCCP->uncompressNext ()) != -1)
  {
    //inform plug-ins about uncompressed data (can be equal to raw data if MCCP isn't used
    invokeEvent ("raw-data", sess(), QString (data));

#if 0
printf ("DATA IN: ");
for (int i = 0; i < datalen; ++i) printf ("%02x ", (unsigned char) data[i]);
printf ("\n");
#endif

    data[datalen] = '\0';
    QString cleandata;

    //we have some data...
    d->newdata = true;

    //clear the GO-AHEAD flag
    d->recvdGA = false;
    
    //now we have the data, but we cannot forward it to next stage of processing,
    //because the data contains telnet commands
    //so we parse the text and process all telnet commands:

    for (unsigned int i = 0; i < (unsigned int) datalen; i++)
    {
      unsigned char ch = data[i];
      if (d->iac || d->iac2 || d->insb || (ch == TN_IAC))
      {
        //there are many possibilities here:
        //1. this is IAC, previous character was regular data
        if (! (d->iac || d->iac2 || d->insb) && (ch == TN_IAC))
        {
          d->iac = true;
          d->command += ch;
        }
        else
        //2. seq. of two IACs
          if (d->iac && (ch == TN_IAC) && (!d->insb))
        {
          d->iac = false;
          char str[2];
          str[0] = ch;
          str[1] = 0;
          cleandata += QString::fromLocal8Bit (str);
          d->command = "";
        }
        else
        //3. IAC DO/DONT/WILL/WONT
          if (d->iac && (!d->insb) &&
            ((ch == TN_WILL) || (ch == TN_WONT) || (ch == TN_DO) || (ch == TN_DONT)))
        {
          d->iac = false;
          d->iac2 = true;
          d->command += ch;
        }
        else
        //4. IAC DO/DONT/WILL/WONT <command code>
          if (d->iac2)
        {
          d->iac2 = false;
          d->command += ch;
          processTelnetCommand (d->command);
          d->command = "";
        }
        else
        //5. IAC SB
          if (d->iac && (!d->insb) && (ch == TN_SB))
        {
          d->iac = false;
          d->insb = true;
          d->command += ch;
        }
        else
        //6. IAC SE without IAC SB - error - ignored
          if (d->iac && (!d->insb) && (ch == TN_SE))
        {
          d->command = "";
          d->iac = false;
        }
        else
        //7. inside IAC SB
          if (d->insb)
        {
          d->command += ch;
          if (d->iac && (ch == TN_SE))  //IAC SE - end of subcommand
          {
            processTelnetCommand (d->command);
            d->command = "";
            d->iac = false;
            d->insb = false;
          }
          if (d->iac)
            d->iac = false;
          else
          if (ch == TN_IAC)
              d->iac = true;
        }
        else
        //8. IAC fol. by something else than IAC, SB, SE, DO, DONT, WILL, WONT
        {
          d->iac = false;
          d->command += ch;
          processTelnetCommand (d->command);
          //this could have set receivedGA to true; we'll handle that later
          // (at the end of this function)
          d->command = "";
        }
      }
      else   //plaintext
      {
        //everything except CRLF is okay; CRLF is replaced by LF(\n) (CR ignored)
        if (ch != 13)
        {
          char str[2];
          str[0] = ch;
          str[1] = 0;
          cleandata += QString::fromLocal8Bit (str);
        }
      }

      //we've just received the GA signal - higher layers shall be informed about it
      if (d->recvdGA)
      {
        //ask MSP parser to look for MSP tags if MSP is enabled
#ifdef HAVE_SOUND
        //look if the user want to have MSP all the time
        bool alwaysmsp = false;
        if (cp)
          alwaysmsp = cp->alwaysMSP();
        //hand data to MSP parser if desired
        if (d->usingmsp || alwaysmsp)
          cleandata = d->MSP->parseServerOutput (cleandata);
#endif

        //prepend a newline, if needed
        if (d->prependGANewLine && d->t_lpmudstyle)
          cleandata = "\n" + cleandata;
        d->prependGANewLine = false;
        //forward data for further processing
        invokeEvent ("data-received", sess(), cleandata);

        if (cp && cp->consolePrompt())
          //we'll need to prepend a new-line in next data sending
          d->prependGANewLine = true;
        //we got a prompt
        invokeEvent ("received-ga", sess());

        //clean the flag, and the data (so that we don't send it multiple times)
        cleandata = QString::null;
        d->recvdGA = false;
      }
    }

    //some data left to send - do it now!
    if (!cleandata.isEmpty())
    {
      //ask MSP parser to look for MSP tags if MSP is enabled
#ifdef HAVE_SOUND
      //look if the user want to have MSP all the time
      bool alwaysmsp = false;
      if (cp)
        alwaysmsp = cp->alwaysMSP();
      //hand data to MSP parser if desired
      if (d->usingmsp || alwaysmsp)
        cleandata = d->MSP->parseServerOutput (cleandata);
#endif

      //prepend a newline, if needed
      if (d->prependGANewLine && d->t_lpmudstyle)
        cleandata = "\n" + cleandata;
      d->prependGANewLine = false;
      //forward data for further processing
      invokeEvent ("data-received", sess(), cleandata);
    }
  }

  invokeEvent ("text-here", sess());
}

void cTelnet::socketClosed ()
{
  disconnect ();
  (KApplication::kApplication())->processEvents ();
}

bool cTelnet::isOffLineConnection () { 
  return d->offLineConnection;
}

int cTelnet::compressedBytes () {
  return d->MCCP->compressedBytes ();
}

int cTelnet::uncompressedBytes () {
  return d->MCCP->uncompressedBytes ();
}

int cTelnet::sentBytes () {
  return d->sentbytes;
}

bool cTelnet::usingMCCP () {
  return d->MCCP->usingMCCP();
}

int cTelnet::MCCPVer () {
  return d->MCCP->MCCPVer();
}

void cTelnet::setQuitCommand (QString cmd)
{
  d->quitcommand = cmd;
}

void cTelnet::setCommandEcho (bool cmdEcho)
{
  d->t_cmdEcho = cmdEcho;
}

void cTelnet::setLPMudStyle (bool lpmudstyle)
{
  d->t_lpmudstyle = lpmudstyle;
}

void cTelnet::setNegotiateOnStartup (bool startupneg)
{
  d->_startupneg = startupneg;
}

