/***************************************************************************
                          ctranscript.cpp  -  session transcript
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Thu oct 31 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ctranscript.h"

#include "cansiparser.h"
#include "cconnprefs.h"
#include "cconsole.h"   // needed for buffer dump
#include "coutput.h"
#include "ctelnet.h"
#include "ctextchunk.h"
#include "kmuddy.h"
#include "dialogs/dlgtranscript.h"
#include "dialogs/dlgdumpbuffer.h"

#include <errno.h>

#include <qdir.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kconfig.h>
#include <kstandarddirs.h>

extern KMuddy *kmuddy;

cTranscript::cTranscript (int sess) : cActionBase ("transcript", sess)
{
  running = false;
  advrunning = false;
  rotatedaily = false;
  file = NULL;
  advfile = NULL;
  overwrite = true;
  includedump = false;
  includetimestamp = false;
  type = advtype = 1;
  nextLineGagged = false;
  
  addEventHandler ("disconnected", 20, PT_NOTHING);
  addEventHandler ("got-line-after-triggers", 50, PT_TEXTCHUNK);
  addEventHandler ("will-gag", 50, PT_NOTHING);
  addEventHandler ("displayed-line", 50, PT_TEXTCHUNK);
  addEventHandler ("displayed-prompt", 50, PT_TEXTCHUNK);
}

cTranscript::~cTranscript ()
{ 
  stopTranscript ();
  stopAdvTranscript ();

  removeEventHandler ("disconnected");
  removeEventHandler ("will-gag");
  removeEventHandler ("got-line-after-triggers");
  removeEventHandler ("displayed-line");
  removeEventHandler ("displayed-prompt");
}

void cTranscript::eventHandler (QString event, int)
{
  if (event == "disconnected") {
    stopTranscript ();
    stopAdvTranscript ();
  }
  else if (event == "will-gag") {
    nextLineGagged = true;
  }
}

void cTranscript::eventHandler (QString event, int, cTextChunk *chunk)
{
  if (event == "displayed-line") {
    addLineToTranscript (chunk);
    addLineToAdvTranscript (chunk);
  }
  else if (event == "displayed-prompt") {
    addLineToTranscript (chunk);
    addLineToAdvTranscript (chunk);
  }
  else if (event == "got-line-after-triggers") {
    // got-line-after-triggers is called for every line, gagged or not
    // if the line will be gagged, will-gag was invoked immediately before this
    if (nextLineGagged) {
      // gagged lines will also be added to the transcript
      // we don't use this event for everything, so that later on, we can configure, whether
      // we want to include gagged lines or not
      addLineToTranscript (chunk);
      addLineToAdvTranscript (chunk);
      nextLineGagged = false;
    }
  }
}

void cTranscript::addLineToTranscript (cTextChunk *chunk)
{
  if (!running)
    return;
  
  cANSIParser *ap = dynamic_cast<cANSIParser *>(object ("ansiparser"));
  QString s;
  switch (type) {
    case 1: s = chunk->toText (); break;
    case 2: s = chunk->toAnsi (ap); break;
    case 3: s = chunk->toHTML (); break;
  };
  QCString ch = s.local8Bit ();
  if (ch)
    fputs (ch, file);
  fflush (file);
}

void cTranscript::addLineToAdvTranscript (cTextChunk *chunk)
{
  if (!advrunning)
    return;
  
  QString timestamp;
  if (includetimestamp)
  {
    QTime time = QTime::currentTime ();
    timestamp.sprintf ("[%02d:%02d:%02d.%02d] ", time.hour(), time.minute(), time.second(), time.msec() / 10);
    fputs (timestamp.latin1(), advfile);
  }
 
  cANSIParser *ap = dynamic_cast<cANSIParser *>(object ("ansiparser"));
  QString s;
  switch (advtype) {
    case 1: s = chunk->toText (); break;
    case 2: s = chunk->toAnsi (ap); break;
    case 3: s = chunk->toHTML (); break;
  };
  QCString ch = s.local8Bit ();
  if (ch)
    fputs (ch, advfile);
  
  fflush (advfile);
}

void cTranscript::stopTranscript ()
{
  if (!running)
    return;
    
  fputs ("\n\n", file);
  
  if (type == 2)  //ANSI transcript
  {
    //set output color to default (usually gray on black)
    char defcolor[5];
    defcolor[0] = 27;
    defcolor[1] = '[';
    defcolor[2] = '0';
    defcolor[3] = 'm';
    defcolor[4] = 0;
    fputs (defcolor, file);
  }
  if (type == 3)  //HTML transcript
  {
    //closing HTML tags
    fputs ("</pre></body></html>\n", file);
  }
  
  running = false;
  fclose (file);
  file = 0;
  invokeEvent ("message", sess(), i18n ("Session transcript has been stopped."));
}

void cTranscript::stopAdvTranscript ()
{
  if (!advrunning)
    return;
    
  fputs ("\n\n", advfile);
  
  if (type == 2)  //ANSI transcript
  {
    //set output color to default (usually gray on black)
    char defcolor[5];
    defcolor[0] = 27;
    defcolor[1] = '[';
    defcolor[2] = '0';
    defcolor[3] = 'm';
    defcolor[4] = 0;
    fputs (defcolor, advfile);
  }
  if (type == 3)  //HTML transcript
  {
    //closing HTML tags
    fputs ("</pre></body></html>\n", advfile);
  }
    
  advrunning = false;
  fclose (advfile);
  advfile = 0;
  if (transtimer->isActive ())
  {
    transtimer->stop ();
    delete transtimer;
  }
  invokeEvent ("message", sess(), i18n ("Advanced session transcript has been stopped."));
}

void cTranscript::configure ()
{
  //so first we have to create the dialog...
  tdlg = new dlgTranscript (kmuddy);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (tdlg, SIGNAL (okClicked()), this, SLOT (applyTranscript ()));
  connect (tdlg, SIGNAL (applyClicked()), this, SLOT (applyTranscript ()));

  //next we fill in its data
  tdlg->setEnabled (running);

  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));
  if (cp)
    fname = cp->transcriptDir();
  else
    fname = QDir::homeDirPath();
  
  fname += "/kmuddy_transcript.txt";

  tdlg->setTranscriptType (type);
  tdlg->setFName (fname);
  tdlg->setAFName (fileformat);
  tdlg->setOverwrite (overwrite);
  tdlg->setIncludeDump (includedump);
  tdlg->setAdvTranscript(advrunning);
  tdlg->setRotateDaily (rotatedaily);
  tdlg->setIncludeTimestamp(includetimestamp);  
  tdlg->setAdvTranscriptType (advtype);
  
  //dialog is ready - show it!
  tdlg->exec ();

  //finally, destroy that dialog!
  delete tdlg;
}

void cTranscript::startTranscript ()
{
  //we'll have multiple <html></html> sections in a transcript file if we continue an existing
  //one, but every browser that I've tested can handle it correctly :D
  if (running)
    stopTranscript ();
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet"));
  if (!(telnet->isConnected()))  //no transcript if we aren't connected
    return;
  errno = 0;
  
  file = fopen (fname.latin1(), overwrite ? "w" : "a");
  if (file == NULL)
  {
    KMessageBox::detailedSorry (kmuddy,
        i18n ("Transcript file could not be opened."), strerror (errno));
    errno = 0;
    invokeEvent ("message", sess(), i18n ("Session transcript could not be started."));
  }
  else
  {
    running = true;
    cOutput *output = dynamic_cast<cOutput *>(object ("output"));
    fputs ("\n\n", file);
    if (type == 3)
    {
      //TODO: what if we're adding to an existing HTML transcript?
      fputs ("<html>\n", file);
      fputs ("<meta name=\"Generator\" content=\"KMuddy\">\n", file);
      fputs ("<body bgcolor=", file);
      fputs (output->defaultBkColor().name().latin1(), file);
      fputs (">\n", file);
    }
    QString s = i18n ("Session transcript has just started.");
    fputs (s.latin1(), file);
    if (type == 3) fputs ("<br><pre>", file);  //pre-formatted text starting...
    fputs ("\n\n", file);

    //add buffer dump if requested
    if (includedump)
      output->console()->dumpBuffer (false, file, type);

    invokeEvent ("message", sess(), i18n ("Session transcript has been started."));
  }
}

void cTranscript::startAdvTranscript ()
{
  QString af;
  
  if (advrunning)
    stopAdvTranscript ();
  cTelnet *telnet = dynamic_cast<cTelnet *>(object ("telnet"));
  if (!(telnet->isConnected()))  //no transcript if we aren't connected
    return;
  
  if(advfname.length() == 0 || !advfname)
    return; // no transcript if advanced filename failed
  
  errno = 0;
 
  transtimer = new QTimer();
  connect (transtimer, SIGNAL (timeout ()), this, SLOT (timeout ()));

  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));
  if (cp)
    af = cp->transcriptDir() + "/" + advfname;
  else
    af = QDir::homeDirPath() + "/" + advfname;

  advfile = fopen (af.latin1(), "a");
  if (advfile == NULL)
  {
    KMessageBox::detailedSorry (kmuddy,
        i18n ("Advanced transcript file could not be opened."), strerror (errno));
    errno = 0;
    invokeEvent ("message", sess(), i18n ("Advanced session transcript could not be started."));
  }
  else
  {
    advrunning = true;
    cOutput *output = dynamic_cast<cOutput *>(object ("output"));
    fputs ("\n\n", advfile);
    if (advtype == 3)
    {
      //TODO: what if we're adding to an existing HTML transcript?
      fputs ("<html>\n", advfile);
      fputs ("<meta name=\"Generator\" content=\"KMuddy\">\n", advfile);
      fputs ("<body bgcolor=", advfile);
      fputs (output->defaultBkColor().name().latin1(), advfile);
      fputs (">\n", advfile);
    }
    QString s = i18n ("Advanced session transcript has just started.");
    fputs (s.latin1(), advfile);
    if (advtype == 3) fputs ("<br><pre>", advfile);  //pre-formatted text starting...
    fputs ("\n\n", advfile);

    invokeEvent ("message", sess(), i18n ("Advanced session transcript has been started."));
    
    if (rotatedaily)
    {
      invokeEvent ("message", sess(), i18n ("Session transcript will be rotated at midnight."));
      transtimer->start (60000);
    }
  }
}

void cTranscript::load()
{
  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));
  name = cp->name();
  directory = locateLocal ("appdata", "profiles/" + name + "/");
  
  KConfig *config = new KConfig ( directory + "transcript", true);
  
  config->setGroup ("Advanced transcript");
  setEnableRotate (config->readBoolEntry ("Daily rotation", false));
  setPrependTimestamp (config->readBoolEntry("Prepend timestamp", false));
  setAFName (config->readEntry("Advanced logname", QString::null));
  advtype = config->readNumEntry ("Transcript type", 1);
  delete config;
}

void cTranscript::save()
{
  KConfig *config = new KConfig ( directory + "transcript");
  
  config->setGroup ("Advanced transcript");
  config->writeEntry ("Daily rotation", rotatedaily);
  config->writeEntry ("Prepend timestamp", includetimestamp);
  config->writeEntry ("Advanced logname", fileformat);
  config->writeEntry ("Transcript type", advtype);
  delete config;
}

void cTranscript::applyTranscript ()
//handler of OK/Apply in transcript dialog
{
  stopTranscript ();
  stopAdvTranscript();
  overwrite = tdlg->isOverwrite ();
  includedump = tdlg->includeDump ();
  fname = tdlg->getFName ();  
  type = tdlg->transcriptType();
  
  fileformat = tdlg->getAFName(); 
  setAFName (fileformat);
  
  if (tdlg->isEnabled ())
    startTranscript ();  
  
  rotatedaily = tdlg->rotateDaily();
  includetimestamp = tdlg->includeTimestamp();
  advtype = tdlg->advTranscriptType();
  
  if(tdlg->advTranscript())
    startAdvTranscript();
  
  save();
}

void cTranscript::dumpBuffer ()
{
  //so first we have to create the dialog...
  bdlg = new dlgDumpBuffer (kmuddy);

  QString fName;
  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));
  if (cp)
    fName = cp->transcriptDir ();
  else
    fName = QDir::homeDirPath ();
  fName += "/buffer_dump.txt";
  bdlg->setFileName (fName);
  bdlg->setType (1);
  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (bdlg, SIGNAL (okClicked()), this, SLOT (doDumpBuffer ()));

  //dialog is ready - show it!
  bdlg->exec ();

  //finally, destroy that dialog!
  delete bdlg;
}

void cTranscript::doDumpBuffer ()
{
  QString fName = bdlg->fileName();
  int type = bdlg->type();
  FILE *f = fopen (fName.latin1(), "w");
  if (f)
  {
    cOutput *output = dynamic_cast<cOutput *>(object ("output"));
    if (type == 3)
    {
      fputs ("<html>\n", f);
      fputs ("<meta name=\"Generator\" content=\"KMuddy\">\n", f);
      fputs ("<body bgcolor=", f);
      fputs (output->defaultBkColor().name().latin1(), f);
      fputs ("><pre>\n", f);
    }
    output->console()->dumpBuffer (bdlg->curPos(), f, type);
    if (type == 3)
      fputs ("\n</pre></body></html>\n", f);
  }
  else
    KMessageBox::detailedSorry (kmuddy,
        i18n ("Dump file could not be opened."), strerror (errno));
  fclose (f);
}

void cTranscript::timeout()
{
  QTime time = QTime::currentTime();
  
  if((time.hour() == 0) && (time.minute() == 0))
  {
    stopAdvTranscript();
    startAdvTranscript();
  }
}

void cTranscript::setAFName(const QString &what)
{
  QString day, month, year, hour, min;
  QString sessname;

  QDate date = QDate::currentDate();
  QTime time = QTime::currentTime();
  
  fileformat = what;
  
  day.sprintf("%02d", date.day());
  month.sprintf("%02d", date.month());
  year.sprintf("%02d", date.year());
  hour.sprintf("%02d", time.hour());
  min.sprintf("%02d", time.minute());

  cConnPrefs *cp = dynamic_cast<cConnPrefs *>(object ("connprefs"));
  if (cp) sessname = cp->name();
      
  if (what)
  {
    advfname = what;
    advfname.replace("$W", sessname);
    advfname.replace("$D", day);
    advfname.replace("$M", month);
    advfname.replace("$Y", year);
    advfname.replace("$h", hour);
    advfname.replace("$m", min);
    
  }
  else
  {
    if(sessname)
      advfname = sessname + "-" + day + "-" + month + "-" + year + ".log";
    else
      advfname = "unnamed-" + day + "-" + month + "-" + year + ".log";
  }
}

QString cTranscript::getAFName()
{
  return advfname;
}
