/***************************************************************************
                          kmuddy.cpp  -  main class that handles interface
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Pi Jun 14 12:37:51 CEST 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmuddy.h"

#include <config.h>

#include "cactionlist.h"
#include "cactionmanager.h"
#include "ccmdparser.h"
#include "cconnection.h"
#include "cconnectioninfo.h"
#include "cconsole.h"
#include "cgaugebar.h"
#include "cgrouplist.h"
#include "cinputline.h"
#include "cmacromanager.h"
#include "cmenumanager.h"
#include "coutput.h"
#include "cpluginmanager.h"
#include "cprofiles.h"
#include "csession.h"
#include "csessionmanager.h"
#include "cshortcutlist.h"
#include "ctranscript.h"
#include "ctelnet.h"
#include "ctabbar.h"
#include "cwintabbar.h"

#include "cglobalsettings.h"
#include "cmultilineinput.h"

#ifdef HAVE_MXP
#include "cmxpconsole.h"
#endif

#include "dialogs/dlgconnect.h"
#include "dialogs/dlgquickconnect.h"
#include "dialogs/dlgeditprofile.h"
#include "dialogs/dlggrabkey.h"
#include "dialogs/dlggroups.h"
#include "dialogs/dlgimportexportprofile.h"
#include "dialogs/dlgstatistics.h"

#include <stdlib.h>
#include <ctime>

#include <qdesktopwidget.h>
#include <qdockarea.h>
#include <qlayout.h>
#include <qtabbar.h>
#include <qtimer.h>
#include <qwidgetstack.h>
#include <qwidget.h>
#include <qpopupmenu.h>
#include <qevent.h>

#include <kaboutapplication.h>
#include <kaction.h>
#include <kapp.h>
#include <khelpmenu.h>
#include <kinputdialog.h>
#include <klocale.h>
#include <kmenubar.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kshortcut.h>
#include <kstatusbar.h>
#include <ktoolbar.h>
#include <kwin.h>
#include <kdebug.h>

//all the objects; everybody who needs them should declare them as externs
cMultiLineInput *multiline;
cGlobalSettings *globalsettings;

#define ID_CONNECTION 1000
#define ID_EDIT 1001
#define ID_VIEW 1002
#define ID_PROFILE 1003
#define ID_TOOLS 1004
#define ID_SETTINGS 1005
#define ID_HELP 1006

//are the actions ready?
bool actionsReady;

bool KMuddy::goingDown = false;

KMuddy::KMuddy(QWidget *parent, const char *name)
  : KMainWindow(parent, name),
    cActionBase ("kmuddy", 0)
{
  //initialize random number generator (needed by cMSP)
  srand (time (0));

  wantToolbar = false;
  tabbar = 0;
  wintabbar = 0;
  actionsReady = false;
  globalnotify = localnotify = alwaysnotify = false;
  agroups = 0;
  tgroups = 0;
  cdlg = 0;
  qdlg = 0;
  mdlg = 0;
  statdlg = 0;
  grabdlg = 0;
}

KMuddy::~KMuddy()
{
  goingDown = true;
  killObjects ();
}

void KMuddy::eventHandler (QString event, int session)
{
  if (event == "notify-request")
    requestNotify (session);
}

void KMuddy::eventHandler (QString event, int session, cTextChunk *)
{
  if ((event == "displayed-line") || (event == "displayed-prompt"))
    if (alwaysnotify) requestNotify (session);
}

void KMuddy::eventHandler (QString event, int session, QString &par1,
    const QString &par2)
{
  if (event == "focus-change")
    focusChange (par1, par2);
}

// some externs ... these are in macros.cpp
void registerInternalMacros ();
void unregisterInternalMacros ();

void KMuddy::prepareObjects ()
{
  //only run this once!!!
  static bool ready = false;
  if (ready)
    return;
  ready = true;

  // plug-in manager
  cPluginManager::self();  // initialise the plugin manager
  registerInternalMacros ();

  globalsettings = new cGlobalSettings;
  multiline = new cMultiLineInput;

  // action manager
  am = cActionManager::self();
  addEventHandler ("displayed-line", 100, PT_TEXTCHUNK);
  addEventHandler ("displayed-prompt", 100, PT_TEXTCHUNK);
  addEventHandler ("notify-request", 50, PT_NOTHING);
  addEventHandler ("focus-change", 50, PT_STRING);
  
  atoolbar = new KToolBar (this, QMainWindow::Top);
  atoolbar->setCaption ("Action toolbar");
  atoolbar->hide ();

  //Let's create the central widget and friends first!

  QFrame *central = new QFrame (this);
  QVBoxLayout *layout = new QVBoxLayout (central);

  //widget stack
  stack = new QWidgetStack (central);
  central->setMinimumSize (400, 300);

  //tab bar
  tabbar = new cTabBar (central);
  connect (tabbar, SIGNAL (selected (int)), this, SLOT (changeSession (int)));
  connect (tabbar, SIGNAL (closeTab()), this, SLOT (closeTab()));
  connect (tabbar, SIGNAL (reConnect()), this, SLOT (doReconnect()));
  connect (tabbar, SIGNAL (disConnect()), this, SLOT (doDisconnect()));

  // give pointers to session manager
  cSessionManager::self()->setWidgetStack (stack);
  cSessionManager::self()->setTabBar (tabbar);
  

  //prepare the new session
  cSessionManager::self()->addSession (false);

  //this addWidget stuff MUST go AFTER the addSession() command!!!
  //otherwise the tabbar is not shown (it contains no tabs at the
  //time of adding to the layout)
  layout->addWidget (tabbar, 1);
  layout->addWidget (wintabbar, 1);
  layout->addWidget (stack, 1);

  //set up the window
  setCentralWidget (central);

  //--- establish connections between objects ---
  //---  most of stuff has moved to cSession  ---

  connect (multiline, SIGNAL (dialogClosed ()), this, SLOT (uncheckMultiLine ()));

  KApplication::kApplication()->processEvents ();

  // now we create the menubar
  cMenuManager::self()->setMenuBar (menuBar());
  
  //get action collection
  cActionManager::self()->createACol (this);
  KActionCollection *acol = cActionManager::self()->getACol ();

  //Connection
  stdconnect = new KAction (i18n ("&Connect..."), "connect_no", CTRL+Key_N, this,
      SLOT(showAndHandleConnectDialog()), acol, "Connect");
  qconnect = new KAction (i18n ("&QuickConnect..."), "connect_creating", CTRL+Key_O, this,
      SLOT (showAndHandleQuickConnectDialog ()), acol, "QuickConnect");
  disconnect = new KAction (i18n ("&Disconnect"), "cancel", CTRL+Key_D, this,
      SLOT (doDisconnect ()), acol, "Disconnect");
  reconnect = new KAction (i18n ("&Reconnect"), CTRL+Key_R, this,
      SLOT (doReconnect ()), acol, "Reconnect");
  closetab = new KAction (i18n ("Close &tab"), KShortcut::null(), this,
      SLOT (closeTab ()), acol, "CloseTab");
  closetab->setEnabled (false);
  reconnect->setEnabled (false);

  //Edit
  pastemenu = new KActionMenu ("Paste &As", "editpaste", this, "PasteMenu");
  acol->insert(pastemenu); //needs manual insert (no option on initialization)
  //"Paste As" items
  pastecommand = new KAction (i18n ("&Command"), KShortcut::null(), this,
      SLOT (pasteCommand()), acol, "PasteCommand");
  pasteinput = new KAction (i18n ("&Input"), KShortcut::null(), this,
      SLOT (pasteInput()), acol, "PasteInput");
  pastemenu->insert(pastecommand);
  pastemenu->insert(pasteinput);

  clipcopy = new KAction (i18n ("Copy &Selection"), "editcopy", CTRL+SHIFT+Key_S, this,
      SLOT (addSelectionToClipboard ()), acol, "ClipboardCopy");
  parsing = new KToggleAction (i18n ("Enable command &parsing"), "goto", CTRL+SHIFT+Key_P,
      acol, "EnableParsing");
  connect (parsing, SIGNAL (toggled (bool)), this, SLOT (setParsing (bool)));
  parsing->setChecked (true);  //enabled by default
  cancelpending = new KAction (i18n ("Ca&ncel pending commands"), "remove", KShortcut::null(),
      this, SLOT (clearCommandQueue ()), acol, "CancelPending");

  //View
  showmultiline = new KToggleAction (i18n ("Show &multi-line input"), "leftjust", CTRL+SHIFT+Key_M,
      multiline, SLOT (toggle ()), acol, "ShowMultiLine");
  showatoolbar = new KToggleAction (i18n ("Show a&ction toolbar"), KShortcut::null(),
      acol, "ShowAToolbar");
  connect (showatoolbar, SIGNAL (toggled (bool)), this, SLOT (showAToolbar (bool)));
  connect (showatoolbar, SIGNAL (toggled (bool)), this, SLOT (setSettingsDirty()));
  showatoolbar->setChecked (false);
  connect (atoolbar, SIGNAL (visibilityChanged (bool)), showatoolbar, SLOT (setChecked (bool)));
  showaliasgroups = new KToggleAction (i18n ("Show &alias groups"), KShortcut::null(),
      acol, "ShowAliasGroups");
  connect (showaliasgroups, SIGNAL (toggled (bool)), this, SLOT (showAliasGroups (bool)));
  showaliasgroups->setChecked (false);
  showtriggergroups = new KToggleAction (i18n ("Show &trigger groups"), KShortcut::null(),
      acol, "ShowTriggerGroups");
  connect (showtriggergroups, SIGNAL (toggled (bool)), this, SLOT (showTriggerGroups (bool)));
  showtriggergroups->setChecked (false);
  showgaugebar = new KToggleAction (i18n ("Show &gaugebar"), KShortcut::null(),
      acol, "ShowGaugeBar");
  connect (showgaugebar, SIGNAL (toggled (bool)), this, SLOT (showGauges (bool)));
  showgaugebar->setChecked (false);
  showstatistics = new KToggleAction (i18n ("Show connection &statistics"), KShortcut::null(),
      acol, "ShowStatistics");
  connect (showstatistics, SIGNAL (toggled (bool)), this, SLOT (showStatistics (bool)));
  showstatistics->setChecked (false);
#ifdef HAVE_MXP
  showmxpconsole = new KToggleAction (i18n ("Show MXP console"), KShortcut::null(),
      acol, "ShowMXPConsole");
  connect (showmxpconsole, SIGNAL (toggled (bool)), this, SLOT (switchMXPConsole (bool)));
  showmxpconsole->setChecked (false);
  //uncheck action when user closes MXP console
  connect (cMXPConsole::self(), SIGNAL (finished()), this, SLOT (MXPConsoleFinished ()));
#endif

  //Profile
  aliases = new KAction (i18n ("&Aliases..."), CTRL+ALT+Key_A, this,
      SLOT (handleAliasesDialog()), acol, "Aliases");
  triggers = new KAction (i18n ("&Triggers..."), CTRL+ALT+Key_T, this,
      SLOT (handleTriggersDialog()), acol, "Triggers");
  actions = new KAction (i18n ("A&ctions..."), KShortcut::null(), this,
      SLOT (handleActionsDialog()), acol, "Actions");
  timers = new KAction (i18n ("Time&rs..."), KShortcut::null(), this,
      SLOT (handleTimersDialog()), acol, "Timers");
  shortcuts = new KAction (i18n ("&Macro keys..."), KShortcut::null(), this,
      SLOT (handleShortcutsDialog()), acol, "MacroKeys");
  variables = new KAction (i18n ("&Variables..."), KShortcut::null(), this,
      SLOT (handleVariablesDialog()), acol, "Variables");
  vartrigs = new KAction (i18n ("Varia&ble triggers..."), KShortcut::null(), this,
      SLOT (handleVarTrigsDialog()), acol, "VarTrigs");
  statusvars = new KAction (i18n ("Status va&riables..."), KShortcut::null(), this,
      SLOT (handleStatusVarsDialog()), acol, "StatusVars");
  gauges = new KAction (i18n ("&Gauges..."), KShortcut::null(), this,
      SLOT (handleGaugesDialog()), acol, "Gauges");
  windows = new KAction (i18n ("&Output windows..."), KShortcut::null(), this,
      SLOT (handleWindowsDialog()), acol, "OutputWindows");
  ticktimer = new KAction (i18n ("T&ick timer..."), KShortcut::null(), this,
      SLOT (handleTickTimerDialog()), acol, "TickTimer");
  switchaliases = new KToggleAction (i18n ("A&liases enabled"), "aliases", CTRL+SHIFT+Key_A,
      acol, "EnableAliases");
  switchtriggers = new KToggleAction (i18n ("Tri&ggers enabled"), "text", CTRL+SHIFT+Key_T,
      acol, "EnableTriggers");
  switchtimers = new KToggleAction (i18n ("Tim&ers enabled"), KShortcut::null(),
      acol, "EnableTimers");
  switchshortcuts = new KToggleAction (i18n ("Macro &keys enabled"), KShortcut::null(),
      acol, "EnableMacroKeys");
  connect (switchaliases, SIGNAL (toggled (bool)), this, SLOT (switchAliases (bool)));
  connect (switchtriggers, SIGNAL (toggled (bool)), this, SLOT (switchTriggers (bool)));
  connect (switchtimers, SIGNAL (toggled (bool)), this, SLOT (switchTimers (bool)));
  connect (switchshortcuts, SIGNAL (toggled (bool)), this, SLOT (switchShortcuts (bool)));
  switchaliases->setChecked (true);
  switchtriggers->setChecked (true);
  switchtimers->setChecked (true);
  switchshortcuts->setChecked (true);
  connprefs = new KAction (i18n ("&MUD Preferences..."), "identity", KShortcut::null(), this,
      SLOT (showConnPrefsDialog()), acol, "ConnPrefs");

  //Tools
  sesstranscript = new KAction (i18n ("&Session transcript..."), "toggle_log", KShortcut::null(),
      this, SLOT (configureTranscript()), acol, "Transcript");
  dumpbuffer = new KAction (i18n ("Dump output &buffer..."), KShortcut::null(),
      this, SLOT (dumpBuffer()), acol, "DumpBuffer");
  decide = new KAction (i18n ("&Decision assistant"), KShortcut::null(),
      this, SLOT (makeDecision()), acol, "Decision");
  importprofile = new KAction (i18n ("&Import profile..."),
      KShortcut::null(), this, SLOT (importProfile()), acol, "ImportProf");
  exportprofile = new KAction (i18n ("&Export profile..."),
      KShortcut::null(), this, SLOT (exportProfile()), acol, "ExportProf");

  //Settings
  appconfig = new KAction (i18n ("&Global settings..."), "configure", KShortcut::null(), globalsettings,
      SLOT(showSettingsDialog()), acol, "Global settings");
  pluginsconfig = new KAction (i18n ("&Plugins..."), "plugins", KShortcut::null(),
      cPluginManager::self(), SLOT(showPluginsDialog()) , acol, "Plugins");
  showmenubar = KStdAction::showMenubar(this, SLOT(toggleShowMenu()),
      acol, "ShowMenubar");
  fullscreenmode = new KToggleAction (i18n ("F&ull screen mode"),
      "window_fullscreen", CTRL+SHIFT+Key_F,  acol, "SetFullScreen");
  connect (fullscreenmode, SIGNAL (toggled (bool)), this,
      SLOT (setFullScreen(bool)));

  //Help

  //other actions
  lineup = new KAction (i18n ("Shift line up"), SHIFT+Key_Up, this,
      SLOT (lineUp ()), acol, "LineUp");
  linedown = new KAction (i18n ("Shift line down"), SHIFT+Key_Down, this,
      SLOT (lineDown ()), acol, "LineDown");
  pageup = new KAction (i18n ("Shift page up"), SHIFT+Key_Prior, this,
      SLOT (pageUp ()), acol, "PageUp");
  pagedown = new KAction (i18n ("Shift page down"), SHIFT+Key_Next, this,
      SLOT (pageDown ()), acol, "PageDown");
  aconup = new KAction (i18n ("Split-screen up"), CTRL+ALT+Key_Up, this,
      SLOT (aconUp ()), acol, "AconUp");
  acondown = new KAction (i18n ("Split-screen down"), CTRL+ALT+Key_Down, this,
      SLOT (aconDown ()), acol, "AconDown");
  prevtab = new KAction (i18n ("Previous tab"), ALT+Key_Prior, this,
      SLOT (prevTab ()), acol, "PreviousTab");
  nexttab = new KAction (i18n ("Next tab"), ALT+Key_Next, this,
      SLOT (nextTab ()), acol, "NextTab");
  tabnum[0] = new KAction (i18n ("Switch to tab 1"), ALT+Key_1, this,
      SLOT (switchTab1 ()), acol, "SwitchTab1");
  tabnum[1] = new KAction (i18n ("Switch to tab 2"), ALT+Key_2, this,
      SLOT (switchTab2 ()), acol, "SwitchTab2");
  tabnum[2] = new KAction (i18n ("Switch to tab 3"), ALT+Key_3, this,
      SLOT (switchTab3 ()), acol, "SwitchTab3");
  tabnum[3] = new KAction (i18n ("Switch to tab 4"), ALT+Key_4, this,
      SLOT (switchTab4 ()), acol, "SwitchTab4");
  tabnum[4] = new KAction (i18n ("Switch to tab 5"), ALT+Key_5, this,
      SLOT (switchTab5 ()), acol, "SwitchTab5");
  tabnum[5] = new KAction (i18n ("Switch to tab 6"), ALT+Key_6, this,
      SLOT (switchTab6 ()), acol, "SwitchTab6");
  tabnum[6] = new KAction (i18n ("Switch to tab 7"), ALT+Key_7, this,
      SLOT (switchTab7 ()), acol, "SwitchTab7");
  tabnum[7] = new KAction (i18n ("Switch to tab 8"), ALT+Key_8, this,
      SLOT (switchTab8 ()), acol, "SwitchTab8");
  tabnum[8] = new KAction (i18n ("Switch to tab 9"), ALT+Key_9, this,
      SLOT (switchTab9 ()), acol, "SwitchTab9");
  tabnum[9] = new KAction (i18n ("Switch to tab 10"), ALT+Key_0, this,
      SLOT (switchTab10 ()), acol, "SwitchTab10");

  actionsReady = true;

  cMenuManager *menu = cMenuManager::self();

  //create all the menus
  connectionMenu = new QPopupMenu (this);
  editMenu = new QPopupMenu (this);
  viewMenu = new QPopupMenu (this);
  profileMenu = new QPopupMenu (this);
  toolsMenu = new QPopupMenu (this);
  settingsMenu = new QPopupMenu (this);
  helpMenu = new KHelpMenu (this, QString::null, false);
  
  menu->addMenuPosition ("menu-base");
  menu->addMenuPosition ("menu-plugin");
  menu->addMenuPosition ("menu-settings");
  menu->addMenuPosition ("menu-help");

  menu->addMenu (connectionMenu, i18n ("&Connection"), "menu-base", ID_CONNECTION);
  menu->addMenu (editMenu, i18n ("&Edit"), "menu-base", ID_EDIT);
  menu->addMenu (viewMenu, i18n ("&View"), "menu-base", ID_VIEW);
  menu->addMenu (profileMenu, i18n ("&Profile"), "menu-base", ID_PROFILE);
  menu->addMenu (toolsMenu, i18n ("&Tools"), "menu-base", ID_TOOLS);
  menu->addMenu (settingsMenu, i18n ("&Settings"), "menu-settings", ID_SETTINGS);
  menu->addMenu (helpMenu->menu(), i18n ("&Help"), "menu-help", ID_HELP);

  // create item slots and add entries to them

  //menu Connection
  menu->addItemPosition ("connection-slot1", connectionMenu);
  menu->addItemPosition ("connection-slot2", connectionMenu);
  menu->addItemPosition ("connection-quit", connectionMenu);
  
  menu->plug (stdconnect, "connection-slot1");
  menu->plug (qconnect, "connection-slot1");
  menu->plug (disconnect, "connection-slot1");
  menu->plug (reconnect, "connection-slot2");
  menu->plug (closetab, "connection-slot2");
  menu->plug (KStdAction::quit(this, SLOT(quit()), 0), "connection-quit");

  //menu Edit
  menu->addItemPosition ("edit-slot1", editMenu);
  menu->addItemPosition ("edit-slot2", editMenu);
  menu->addItemPosition ("edit-slot3", editMenu);
  menu->plug (clipcopy, "edit-slot1");
  menu->plug (pastemenu, "edit-slot1");
  menu->plug (parsing, "edit-slot2");
  menu->plug (cancelpending, "edit-slot3");
  
  //menu View
  menu->addItemPosition ("view-global", viewMenu);
  menu->addItemPosition ("view-profile", viewMenu);
  
  menu->plug (showmultiline, "view-global");
  menu->plug (showstatistics, "view-global");
  menu->plug (showaliasgroups, "view-profile");
  menu->plug (showtriggergroups, "view-profile");
  menu->plug (showatoolbar, "view-profile");
  menu->plug (showgaugebar, "view-profile");
#ifdef HAVE_MXP
  menu->plug (showmxpconsole, "view-profile");
#endif

  //menu Profile
  menu->addItemPosition ("profile-basicobjs", profileMenu);
  menu->addItemPosition ("profile-objects", profileMenu);
  menu->addItemPosition ("profile-prefs", profileMenu);
  menu->addItemPosition ("profile-switches", profileMenu);

  menu->plug (aliases, "profile-basicobjs");
  menu->plug (triggers, "profile-basicobjs");
  menu->plug (variables, "profile-basicobjs");
  menu->plug (actions, "profile-objects");
  menu->plug (timers, "profile-objects");
  menu->plug (shortcuts, "profile-objects");
  menu->plug (vartrigs, "profile-objects");
  menu->plug (statusvars, "profile-objects");
  menu->plug (gauges, "profile-objects");
  menu->plug (windows, "profile-objects");
  menu->plug (ticktimer, "profile-objects");
  menu->plug (connprefs, "profile-prefs");
  menu->plug (switchaliases, "profile-switches");
  menu->plug (switchtriggers, "profile-switches");
  menu->plug (switchtimers, "profile-switches");
  menu->plug (switchshortcuts, "profile-switches");

  //menu Tools
  menu->addItemPosition ("tools-slot1", toolsMenu);
  menu->addItemPosition ("tools-slot2", toolsMenu);
  menu->addItemPosition ("tools-slot3", toolsMenu);
  
  menu->plug (sesstranscript, "tools-slot1");
  menu->plug (dumpbuffer, "tools-slot1");
  menu->plug (decide, "tools-slot2");
  menu->plug (importprofile, "tools-slot3");
  menu->plug (exportprofile, "tools-slot3");

  //menu Settings
  menu->addItemPosition ("settings-slot1", settingsMenu);
  menu->addItemPosition ("settings-slot2", settingsMenu);
  menu->addItemPosition ("settings-slot3", settingsMenu);
  
  menu->plug (showmenubar, "settings-slot1");
  menu->plug (fullscreenmode, "settings-slot2");
  menu->plug (appconfig, "settings-slot3");
  menu->plug (pluginsconfig, "settings-slot3");

  //menu Help
      //make About Kmuddy work (I have no idea why it doesn't work immediately)
  connect (helpMenu, SIGNAL (showAboutApplication ()), this, SLOT (aboutApp ()));

  //read saved keys for acol object
  acol->readShortcutSettings ("Shortcuts");

  // load all the plug-ins
  cPluginManager::self()->loadAll ();

  //we're not connected on startup, so let's disable some options...
  disableConnectedOptions ();
  disableAdvancedOptions ();

  //apply global settings
  globalsettings->apply ();

  //look if there's some auto-connect profile set
  if (autoconnect.length() > 0)
    doProfileConnect (autoconnect, false);

  //let the window remember its position!
  setAutoSaveSettings ();

  //make sure that menu bar is visible each time kmuddy is started
  if (!menuBar()->isVisible())
    toggleShowMenu ();
}

void KMuddy::pasteCommand ()
{
  //Should we allow it to addCommand even with no global clipboard input?
  //If we do, sends a newline (no big deal, may even be a 'feature')
  //All depends on how we want this set up; atm not allowing empty clipboard

  if ((KApplication::kApplication())->clipboard()->text().isNull())
    return;

  QString txt =
    (KApplication::kApplication())->clipboard()->text(QClipboard::Clipboard);
  am->invokeEvent ("send-command", activeSession(), txt);
}

void KMuddy::pasteInput ()
{
  cInputLine *inputline = dynamic_cast<cInputLine *>(am->object ("inputline", activeSession()));
  inputline->paste ();
}

void KMuddy::killObjects ()
{
  //destroy objects and set pointers to 0 - needed so that we can prevent
  //some crashes if some object calls a function of an object that has already
  //been destroyed; this gives him a chance to verify its state...

  //menubar and KActions are not deleted...

  //session objects are NOT deleted here - doing so causes kmuddy to crash

  removeEventHandler ("displayed-line");
  removeEventHandler ("displayed-prompt");
  removeEventHandler ("notify-request");
  removeEventHandler ("focus-change");

  delete multiline;
  multiline = 0;
  delete globalsettings;
  globalsettings = 0;

  delete cPluginManager::self();
  unregisterInternalMacros ();
  delete cMacroManager::self();

  //also delete some dialogs...
  if (agroups != 0)
    delete agroups;
  agroups = 0;
  if (tgroups != 0)
    delete tgroups;
  tgroups = 0;
  if (statdlg != 0)
    delete statdlg;
  statdlg = 0;
}

void KMuddy::saveProperties (KConfig * /*config*/)
{
  //nothing here for now, maybe I'll include re-connecting later?
}

void KMuddy::readProperties (KConfig * /*config*/)
{
  //nothing here now now, maybe I'll include re-connecting later?
}

void KMuddy::quit ()
{
  for ( int i = 1; i <= MAXSESSIONS; i++ ) {
    if (am->sessionExists (i)) {
      cConnection *conn = dynamic_cast<cConnection *>(am->object ("connection", i));
      if ( !conn->handleDisconnect() ) {
        return;
      }
    }
  }

  kapp->quit();
}

void KMuddy::disableConnectedOptions ()
//may contain further commands...
{
  if (!actionsReady) return;

  disconnect->setEnabled (false);
  sesstranscript->setEnabled (false);
  dumpbuffer->setEnabled (false);

  disableAdvancedOptions ();
}

void KMuddy::enableConnectedOptions ()
//may contain further commands...
{
  if (!actionsReady) return;
  
  if (!(disconnect->isEnabled()))
    disconnect->setEnabled (true);
  if (!(sesstranscript->isEnabled()))
    sesstranscript->setEnabled (true);
  if (!(dumpbuffer->isEnabled()))
    dumpbuffer->setEnabled (true);
  reconnect->setEnabled (false);
}

void KMuddy::disableAdvancedOptions ()
//may contain further commands...
{
  if (!actionsReady) return;
  
  doShowAToolbar (false);
  showatoolbar->setEnabled (false);
  showgaugebar->setEnabled (false);
  menuBar()->setItemEnabled (ID_PROFILE, false);

  //!!! must also disable all profile-related actions!!! (otherwise
  //they're still accessible with hotkeys)
  aliases->setEnabled (false);
  triggers->setEnabled (false);
  actions->setEnabled (false);
  timers->setEnabled (false);
  shortcuts->setEnabled (false);
  variables->setEnabled (false);
  vartrigs->setEnabled (false);
  statusvars->setEnabled (false);
  gauges->setEnabled (false);
  windows->setEnabled (false);
  ticktimer->setEnabled (false);
  switchaliases->setEnabled (false);
  switchtriggers->setEnabled (false);
  switchtimers->setEnabled (false);
  switchshortcuts->setEnabled (false);
  connprefs->setEnabled (false);
}

void KMuddy::enableAdvancedOptions ()
//may contain further commands...
{
  if (!actionsReady) return;
  
  showatoolbar->setEnabled (true);
  showgaugebar->setEnabled (true);
  menuBar()->setItemEnabled (ID_PROFILE, true);

  //show the toolbar if we want it
  if (wantToolbar)
  {
    doShowAToolbar (true);
    cActionList *al = dynamic_cast<cActionList *>(am->object ("actionlist", activeSession()));
    if (al) al->rebuildToolbar();
  }

  aliases->setEnabled (true);
  triggers->setEnabled (true);
  actions->setEnabled (true);
  timers->setEnabled (true);
  shortcuts->setEnabled (true);
  variables->setEnabled (true);
  vartrigs->setEnabled (true);
  statusvars->setEnabled (true);
  gauges->setEnabled (true);
  windows->setEnabled (true);
  ticktimer->setEnabled (true);
  switchaliases->setEnabled (true);
  switchtriggers->setEnabled (true);
  switchtimers->setEnabled (true);
  switchshortcuts->setEnabled (true);
  connprefs->setEnabled (true);
}

void KMuddy::setClosedConn (bool canCT)
{
  if (!actionsReady) return;
  
  closetab->setEnabled (canCT);
  reconnect->setEnabled (true);
}

void KMuddy::unsetClosedConn ()
{
  if (!actionsReady) return;
  
  closetab->setEnabled (false);
  reconnect->setEnabled (false);
}

void KMuddy::closeTab ()
{
  //close that session...
  cConnection *conn = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));

  if (conn->isConnected ())
    doDisconnect();
  if (!conn->isConnected())
    cSessionManager::self()->removeSession (activeSession(), false);
}

void KMuddy::setFullScreen (bool val)
{
  if (val) {
    showFullScreen();
    if (menuBar()->isVisible())
      toggleShowMenu();
    showmenubar->setChecked( ! menuBar()->isHidden() );
  }
  else
    if (isFullScreen()) {
      showNormal();
      if (!menuBar()->isVisible())
        toggleShowMenu();
      showmenubar->setChecked( ! menuBar()->isHidden() );
    }
}

void KMuddy::toggleShowMenu()
{
  if (menuBar()->isVisible())
    menuBar()->hide();
  else
    menuBar()->show();
}

void KMuddy::doShowAToolbar (bool val)
//show/hide action toolbar
//does not modify wantToolbar variable, so that toolbar can be
//shown/hidden as needed
{
  bool shown = atoolbar->isVisible ();
  if (val)
    atoolbar->show ();
  else
    atoolbar->hide ();

  if (val != shown)
    //showing/hiding toolbar shifts view - fix it!
    //!!! THIS IS FAR FROM BEING IDEAL !!!
    for (int i = 1; i <= MAXSESSIONS; i++)
      if (am->sessionExists (i)) {
        cConsole *console = (dynamic_cast<cOutput *>(am->object
            ("output", activeSession())))->console();
        console->pageDown ();
      }
}

void KMuddy::showAToolbar (bool val)
//Remember this value then call doShowAToolbar
//This is a SLOT. Do NOT call this one. Use doShowAToolbar instead.
{
  if (val)
    wantToolbar = true;
  else
    wantToolbar = false;

  doShowAToolbar (val);

  //keep action in sync (this function can be explicitly called)
  showatoolbar->setChecked (val);
}

void KMuddy::showAliasGroups (bool val)
{
  cGroupList *list;
  list = dynamic_cast<cGroupList *>(am->object ("aliasgrouplist", activeSession()));
  if (!list) return;

  if (agroups == 0)
  {
    agroups = new dlgGroups (list, i18n ("Alias groups"), this);
    connect (agroups, SIGNAL (finished()), this, SLOT (aliasGroupsFinished ()));
  }
  if (val)
    agroups->show ();
  else
    agroups->hide ();

  //keep action in sync (this function can be explicitly called)
  showaliasgroups->setChecked (val);
}

void KMuddy::showTriggerGroups (bool val)
{
  cGroupList *list;
  list = dynamic_cast<cGroupList *>(am->object ("triggergrouplist", activeSession()));
  if (!list) return;

  if (tgroups == 0)
  {
    tgroups = new dlgGroups (list, i18n ("Trigger groups"), this);
    connect (tgroups, SIGNAL (finished()), this, SLOT (triggerGroupsFinished ()));
  }
  if (val)
    tgroups->show ();
  else
    tgroups->hide ();

  //keep action in sync (this function can be explicitly called)
  showtriggergroups->setChecked (val);
}

void KMuddy::showGauges (bool val)
{
  cGaugeBar *gauges = dynamic_cast<cGaugeBar *>(am->object ("gaugebar", activeSession()));

  bool shown = gauges->isVisible ();
  if (val)
    gauges->show ();
  else
    gauges->hide ();

  //keep action in sync (this function can be explicitly called)
  showgaugebar->setChecked (val);

  if (val != shown)
    //showing/hiding toolbar shifts view - fix it!
    //!!! THIS IS FAR FROM BEING IDEAL !!!
    for (int i = 1; i <= MAXSESSIONS; i++)
      if (am->sessionExists (i)) {
        cConsole *console = (dynamic_cast<cOutput *>(am->object
            ("output", activeSession())))->console();
        console->pageDown ();
      }
}

void KMuddy::showStatistics (bool val)
{
  if (statdlg == 0)
  {
    statdlg = new dlgStatistics (this);
    connect (statdlg, SIGNAL (finished()), this, SLOT (statisticsFinished ()));
  }
  if (val)
    statdlg->show ();
  else
    statdlg->hide ();

  //keep action in sync (this function can be explicitly called)
  showstatistics->setChecked (val);
}

void KMuddy::aliasGroupsFinished ()
{
  showAliasGroups (false);
}

void KMuddy::triggerGroupsFinished ()
{
  showTriggerGroups (false);
}

void KMuddy::statisticsFinished ()
{
  showStatistics (false);
}

void KMuddy::switchMXPConsole (bool value)
{
#ifdef HAVE_MXP
  cMXPConsole *mxpc = cMXPConsole::self ();

  if (value)
    mxpc->show ();
  else
    mxpc->hide ();

  //keep action in sync (this function can be explicitly called)
  showmxpconsole->setChecked (value);
#endif
}

void KMuddy::MXPConsoleFinished ()
{
#ifdef HAVE_MXP
  switchMXPConsole (false);
#endif
}

void KMuddy::uncheckMultiLine ()
{
  showmultiline->setChecked (false);
}

// TODO: this should get moved to session manager and command processor
// session manager would return session ID with a given prefix, command
// processor would send command ...
void KMuddy::focusChange (const QString &window, const QString &command)
{
  QString sessName;
  int i=0,id=-1;

  for (i = 1; i <= MAXSESSIONS; i++)
    if (am->sessionExists (i)) {
      sessName = am->callAction ("session", "name", i);
      if ( !sessName.isNull() && !sessName.isEmpty() )
        if (sessName.startsWith(window)) {
          id = i;
          //breaking here means that if more than one session matching
          //window name is found, that sessionName will use the first
          //ie - sessions: 'abc' and 'abc123' - whichever has the lowest
          //sessionid will be the session the command is sent to if only
          //'abc' is passed as the WINDOW
          break;
        }
    }
  if (id <= 0)
    return; //return without sending command

  // send the command to the necessary session - it will be sent as-is
  am->invokeEvent ("send-command", id, command);
}

QString KMuddy::reconnectText ()
{
  QString text = i18n (" To reconnect, press");
  text += " " + reconnect->shortcut().toString() + " ";
  text += i18n ("or use Connection / Reconnect.");
  return text;
}

void KMuddy::aboutApp ()
{
  KAboutApplication *dlg = new KAboutApplication (this);
  dlg->exec ();
  delete dlg;
}

void KMuddy::updateWindows ()
{
  // TODO: the windows could react on events and update themselves ...

  int s = activeSession();
  cGroupList *agl = dynamic_cast<cGroupList *>(am->object ("aliasgrouplist", s));
  cGroupList *tgl = dynamic_cast<cGroupList *>(am->object ("triggergrouplist", s));

  // update item in alias/trigger groups dialog
  // agl,tgl,rl can be 0 - that's okay
  if (agroups != 0)
    agroups->switchGroupList (agl);
  if (tgroups != 0)
    tgroups->switchGroupList (tgl);

  //still not enough; we also need to update the statistics window
  if (statdlg != 0)
    statdlg->update ();
}

void KMuddy::changeSession (int id)
{
  cSessionManager::self()->changeSession (id);
}

int KMuddy::activeSession ()
{
  return cActionManager::self()->activeSession();
}

void KMuddy::setAutoConnect (const QString &ac)
{
  autoconnect = ac;
}

void KMuddy::prevTab ()
{
  int idx = tabbar->indexOf (tabbar->currentTab ());
  idx--;
  if (idx < 0)
    idx = tabbar->count() - 1;
  QTab *tab = tabbar->tabAt (idx);
  if (tab != 0)
    tabbar->setCurrentTab (tab);
}

void KMuddy::nextTab ()
{
  int idx = tabbar->indexOf (tabbar->currentTab ());
  idx++;
  if (idx >= tabbar->count())
    idx = 0;
  QTab *tab = tabbar->tabAt (idx);
  if (tab != 0)
    tabbar->setCurrentTab (tab);
}

void KMuddy::switchTab (int index)
{
  QTab *tab = tabbar->tabAt (index);
  if (tab != 0)   //tab exists
    tabbar->setCurrentTab (tab);
}

void KMuddy::setAllowGlobalNotify (bool notify)
{
  globalnotify = notify;
}

void KMuddy::setAllowLocalNotify (bool notify)
{
  localnotify = notify;
}

void KMuddy::setAlwaysNotify (bool notify)
{
  alwaysnotify = notify;
}

void KMuddy::requestNotify (int sess)
{
  //global notify
  if (globalnotify && (!isActiveWindow ()))
    // ask the window manager to draw attention to the window
    KWin::demandAttention (topLevelWidget()->winId());

  //local notify
  if (localnotify && (activeSession() != sess))
    cSessionManager::self()->setNotifyFlag (sess);
}

void KMuddy::setMenuAliasesEnabled (bool val)
{
  switchaliases->setChecked (val);
}

void KMuddy::setMenuTriggersEnabled (bool val)
{
  switchtriggers->setChecked (val);
}

void KMuddy::setMenuTimersEnabled (bool val)
{
  switchtimers->setChecked (val);
}

void KMuddy::setMenuShortcutsEnabled (bool val)
{
  switchshortcuts->setChecked (val);
}

void KMuddy::doDisconnect ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleDisconnect ();
}

void KMuddy::doReconnect ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->reconnect ();

  // update alias/trigger groups and similar things ...
  updateWindows ();
}

//THESE METHODS WERE ORIGINALLY IN cConnection; now they're here because
//of multi-tab support

void KMuddy::doQuickConnect ()
{
  //hide the dialog so that it doesn't look like we've crashed
  qdlg->hide ();

  //this object will be deleted in cConnection::disconnect()
  cConnectionInfo *i = new cConnectionInfo;
  i->setServer (qdlg->host());
  i->setPort (qdlg->port());

  //create a new session object and related stuff
  int s = cSessionManager::self()->addSession (false);
  cConnection *connection = dynamic_cast<cConnection *>(am->object ("connection", s));
  //now establish that connection!
  connection->establishQuickConnection (i);

  updateWindows ();
}

void KMuddy::doConnect ()
{
  //hide the dialog, so that it doesn't look like we've crashed
  cdlg->hide ();

  QString name = cdlg->selName();
  bool sendNothing = cdlg->sendNothing();

  doProfileConnect (name, sendNothing);
}

void KMuddy::doProfileConnect (const QString &name, bool sendNothing)
{
  cProfiles *prof = new cProfiles;
    //this object will be deleted in cConnection::disconnect()
  cConnectionInfo *i;
  if (prof->exists ((QString &) name))
  {
    i = prof->getProfileConnInfo ((QString &) name);
    delete prof;
  }
  else
  {
    //this will only happen with auto-connecting to a deleted profile;
    //Connect dialog won't allow this
    KMessageBox::sorry (this, i18n ("Sorry, requested profile does not exist."), i18n ("Profile does not exist"));
    delete prof;
    return;
  }

  //create cSession object and related stuff
  int s = cSessionManager::self()->addSession (true);

  // offline connection
  if (cdlg && cdlg->isOffLine()) {
    cTelnet *telnet = dynamic_cast<cTelnet *>(am->object ("telnet", s));
    telnet->setOffLineConnection(true);
  }

  //now establish that connection!
  cConnection *connection = dynamic_cast<cConnection *>(am->object ("connection", s));
  connection->establishConnection (i, sendNothing);

  updateWindows ();
}

void KMuddy::showAndHandleConnectDialog ()
{
  //so first we have to create the dialog...
  cdlg = new dlgConnect (this);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (cdlg, SIGNAL (okClicked()), this, SLOT (doConnect()));
  connect (cdlg, SIGNAL (addPressed()), this, SLOT (addPressed()));
  connect (cdlg, SIGNAL (modifyPressed()), this, SLOT (modifyPressed()));
  connect (cdlg, SIGNAL (deletePressed()), this, SLOT (deletePressed()));
  connect (cdlg, SIGNAL (duplicatePressed()), this, SLOT (duplicatePressed()));

  //dialog is ready - show it!
  cdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete cdlg;
  cdlg = 0;
}

void KMuddy::showAndHandleQuickConnectDialog ()
{
  //so first we have to create the dialog...
  qdlg = new dlgQuickConnect (this);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (qdlg, SIGNAL (okClicked()), this, SLOT (doQuickConnect()));

  //dialog is ready - show it!
  qdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete qdlg;
}

void KMuddy::addPressed ()
{
  //so first we have to create the dialog...
  mdlg = new dlgEditProfile (true, this);

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (mdlg, SIGNAL (okClicked ()), this, SLOT (doAdd ()));

  //fill in default login sequence
  QStringList ls;
  ls.append ("$name");
  ls.append ("$password");
  mdlg->setConnectionString (ls);

  //dialog is ready - show it!
  mdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete mdlg;
}

void KMuddy::doAdd ()
{
  cProfiles *prof = new cProfiles;
  QString name = mdlg->name();
  name = name.simplifyWhiteSpace ();
  cConnectionInfo *cci = new cConnectionInfo;
  cci->setName(name);
  cci->setServer(mdlg->server());
  cci->setPort(mdlg->port());
  cci->setLogin(mdlg->login());
  cci->setPassword(mdlg->password());
  cci->setConnStr(mdlg->connectionString());
  if (prof->exists(name))   //profile already exists
  {
    if (KMessageBox::questionYesNoCancel (this,
        i18n("Profile named") + " " + name + " " + i18n ("already exists.") +
        "\n" + i18n ("Do you want to modify existing profile?"),
        i18n("Add profile")) != KMessageBox::Yes)
      return;
    //they want to modify existing profile, DO IT!
    prof->modifyProfile(name, cci);
  }
  else
    prof->addProfile(name, cci);
  delete prof;
  delete cci;

  cdlg->updateList ();
  cdlg->setSelectedItem (name);
}

void KMuddy::modifyPressed ()
{
  //fetch profile name
  QString name = cdlg->selName();
  if (name == QString::null)
    return;

  //so first we have to create the dialog...
  mdlg = new dlgEditProfile (false, this);

  //then we fill in the data
  cProfiles *prof = new cProfiles;
  cConnectionInfo *cci = prof->getProfileConnInfo (name);
  delete prof;
  mdlg->setName(cci->name());
  mdlg->setServer(cci->server());
  mdlg->setPort(cci->port());
  mdlg->setLogin(cci->login());
  mdlg->setPassword(cci->password());
  mdlg->setConnectionString(cci->connStr());
  delete cci;

  //then we connect() all its signals - this handles everything that the dialog offers...
  connect (mdlg, SIGNAL (okClicked ()), this, SLOT (doModify ()));

  //dialog is ready - show it!
  mdlg->exec ();

  //further action is handled by signal issued by OK button, so that we only
  //have to destroy the dialog...
  delete mdlg;
}

void KMuddy::doModify ()
{
  cProfiles *prof = new cProfiles;
  QString name = mdlg->name();
  name = name.simplifyWhiteSpace ();
  cConnectionInfo * cci = new cConnectionInfo;
  cci->setName(name);
  cci->setServer(mdlg->server());
  cci->setPort(mdlg->port());
  cci->setLogin(mdlg->login());
  cci->setPassword(mdlg->password());
  cci->setConnStr(mdlg->connectionString());
  prof->modifyProfile (name, cci);

  delete prof;
  delete cci;

  cdlg->updateList ();
}

void KMuddy::deletePressed ()
{
  //fetch profile name
  QString name = cdlg->selName();
  if (name == QString::null)
    return;

  //ask if he really wants to do this
  if (KMessageBox::questionYesNoCancel (cdlg,
        i18n ("Do you really want to delete profile " + name + " ?"),
        i18n ("Delete profile")) == KMessageBox::Yes)
  {
    //if he does, DO IT!!!
    cProfiles *prof = new cProfiles ();
    prof->deleteProfile(name);
      //maybe I'll should add some notification if we fail...
    delete prof;
    cdlg->updateList();
  }
}

void KMuddy::duplicatePressed ()
{
  //fetch profile name
  QString name = cdlg->selName();
  if (name == QString::null)
    return;

  //so first we have to ask for a new name...
  bool ok;
  QString newname = KInputDialog::getText (i18n ("Duplicate Profile"), i18n ("Please enter name for the duplicated profile"), QString::null, &ok, cdlg);
  if (ok)
  {
    cProfiles *prof = new cProfiles;
    if (prof->exists (newname))
      KMessageBox::sorry (cdlg, i18n ("Sorry, you have to enter a profile name that does not exist."), i18n ("Profile already exists"));
    else
    {
      prof->duplicateProfile (name, newname);
      //maybe I'll should add some notification if we fail...
      delete prof;
      cdlg->updateList();
    }
  }
}

/** lots of slots that handle menus; were handled by slots in objects,
now we have cSession, so we have to handle it here */
void KMuddy::addSelectionToClipboard ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->addSelectionToClipboard(QClipboard::Clipboard);
}

void KMuddy::setParsing (bool value)
{
  for (int i = 1; i <= MAXSESSIONS; i++)
    if (am->sessionExists (i)) {
      cCmdParser *cmdparser = dynamic_cast<cCmdParser *>(am->object ("cmdparser", i));
      cmdparser->setParsing(value);
    }
}

void KMuddy::clearCommandQueue ()
{
  // NOTHING HERE
  // TODO: make it work when command queues get implemented
  am->invokeEvent ("message", activeSession(), "Feature not implemented yet.");
}

void KMuddy::handleAliasesDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleAliasesDialog();
}

void KMuddy::handleTriggersDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleTriggersDialog();
}

void KMuddy::handleActionsDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleActionsDialog();
}

void KMuddy::handleTimersDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleTimersDialog();
}

void KMuddy::handleShortcutsDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleShortcutsDialog();
}

void KMuddy::handleVariablesDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleVariablesDialog();
}

void KMuddy::handleVarTrigsDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleVarTrigsDialog();
}

void KMuddy::handleStatusVarsDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleStatusVarsDialog();
}

void KMuddy::handleGaugesDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleGaugesDialog();
}

void KMuddy::handleWindowsDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleWindowsDialog();
}

void KMuddy::handleTickTimerDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->handleTickTimerDialog();
}

void KMuddy::switchAliases (bool value)
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->switchAliases(value);
}

void KMuddy::switchTriggers (bool value)
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->switchTriggers(value);
}

void KMuddy::switchTimers (bool value)
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->switchTimers(value);
}

void KMuddy::switchShortcuts (bool value)
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->switchShortcuts(value);
}

void KMuddy::showConnPrefsDialog ()
{
  cConnection *connection;
  connection = dynamic_cast<cConnection *>(am->object ("connection", activeSession()));
  connection->showConnPrefsDialog();
}

void KMuddy::configureTranscript ()
{
  cTranscript *transcript;
  transcript = dynamic_cast<cTranscript *>(am->object ("transcript", activeSession()));
  transcript->configure();
}

void KMuddy::dumpBuffer ()
{
  cTranscript *transcript;
  transcript = dynamic_cast<cTranscript *>(am->object ("transcript", activeSession()));
  transcript->dumpBuffer();
}

void KMuddy::makeDecision ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->makeDecision();
}

void KMuddy::importProfile ()
{
  dlgImportExportProfile *iep = new dlgImportExportProfile (true, this);
  iep->doThings ();
  delete iep;
}

void KMuddy::exportProfile ()
{
  dlgImportExportProfile *iep = new dlgImportExportProfile (false, this);
  iep->doThings ();
  delete iep;
}

void KMuddy::lineUp ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->lineUp();
}

void KMuddy::lineDown ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->lineDown();
}

void KMuddy::pageUp ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->pageUp();
}

void KMuddy::pageDown ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->pageDown();
}

void KMuddy::aconUp ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->aconUp();
}

void KMuddy::aconDown ()
{
  cOutput *output = dynamic_cast<cOutput *>(am->object ("output", activeSession()));
  output->console()->aconDown();
}

bool KMuddy::eventFilter (QObject *o, QEvent *e)
{
  //only handle keypresses, nothing else...
  if (e->type() == QEvent::KeyPress)
  {
    bool report = true;
    QKeyEvent *qke = (QKeyEvent *) e;
    int key = qke->key ();
    int state = qke->state ();

    //do not report modifiers
    switch (key) {
      case Qt::Key_Shift:
      case Qt::Key_Control:
      case Qt::Key_Meta:
      case Qt::Key_Alt:
        report = false;
        break;
      default:
        report = true;
        break;
    };

    if (report)
    {
      if (grabdlg)
      {
        grabdlg->gotKey (key, state);
        return true;
      }
      else
      {
        cShortcutList *sl;
        sl = dynamic_cast<cShortcutList *>(am->object ("shortcutlist", activeSession()));
        if (sl && sl->handleKey (key, state))
              return true;
      }
    }
  }
  return KMainWindow::eventFilter (o, e);
}

bool KMuddy::queryClose ()
{
  for (int i = 1; i <= MAXSESSIONS; i++) {
    if (am->sessionExists (i)) {
      //switch to that session => the user knows which one is about to close
      cSessionManager::self()->setSession (i);
      kapp->processEvents ();
      //disconnect, asking the user whether he really wants to
      cConnection *connection = dynamic_cast<cConnection *>(am->object ("connection", i));
      if (!connection->handleDisconnect())
        return false;
    }
  }

  return true;
}

