//
// Description: All core internal macros and functions.
//     Also includes (un)registerInternalMacros
//
//
// Author: Tomas Mecir <kmuddy@kmuddy.net>, 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "cmacromanager.h"
#include "cactionmanager.h"
#include "ccmdqueue.h"
#include "ccmdprocessor.h"
#include "cgrouplist.h"
#include "ctimerlist.h"
#include "ctimertick.h"
#include "cvariablelist.h"

#include "dialogs/dlggroups.h"

#include <klocale.h>

using namespace std;

class cMacroSet : public cMacro {
public:
  cMacroSet () : cMacro ("set") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    QString value = params.section (' ', 1, -1, QString::SectionSkipEmpty);
    //no variable expansion for our parameter, to allow various tricks :)
    varList(sess)->set (varname, value);
  }
};

class cMacroUnset : public cMacro {
public:
  cMacroUnset () : cMacro ("unset") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    varList(sess)->unset (varname);
  }
};

class cMacroSetVal : public cMacro {
  public:
    cMacroSetVal () : cMacro ("setval") {}
    virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
      // this is like /set, but the value gets parsed using as expression
      cCmdProcessor *cmdproc = dynamic_cast<cCmdProcessor *>(am->object ("cmdprocessor", sess));
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString value = params.section (' ', 1, -1, QString::SectionSkipEmpty);
      bool ok;
      cValue val = cmdproc->eval (value, queue, ok);
      if (!ok) {
        am->invokeEvent ("message", sess, i18n ("/setval: Expression parsing failed."));
        return;
      }
      varList(sess)->set (varname, &val);
    }
};

class cMacroLSet : public cMacro {
  public:
    cMacroLSet () : cMacro ("lset") {}
    virtual void eval (const QString &params, int, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString value = params.section (' ', 1, -1, QString::SectionSkipEmpty);
      //no variable expansion for our parameter, to allow various tricks :)
      queue->setValue (varname, value);
    }
};

class cMacroLUnset : public cMacro {
  public:
    cMacroLUnset () : cMacro ("lunset") {}
    virtual void eval (const QString &params, int, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      queue->delValue (varname);
    }
};

class cMacroInc : public cMacro {
public:
  cMacroInc () : cMacro ("inc") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    QString value = params.section (' ', 1, 1, QString::SectionSkipEmpty);
    bool ok = false;
    double val = value.toDouble(&ok);
    if (ok) varList(sess)->inc (varname, val);
  }
};

class cMacroDec : public cMacro {
public:
  cMacroDec () : cMacro ("dec") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    QString value = params.section (' ', 1, 1, QString::SectionSkipEmpty);
    bool ok = false;
    double val = value.toDouble(&ok);
    if (ok) varList(sess)->dec (varname, val);
  }
};

class cMacroLInc : public cMacro {
  public:
    cMacroLInc () : cMacro ("linc") {}
    virtual void eval (const QString &params, int, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString value = params.section (' ', 1, 1, QString::SectionSkipEmpty);
      bool ok = false;
      double val = value.toDouble(&ok);
      if (ok) {
        cValue *v = queue->value (varname);
        if (!v)  // no such variable ?
          queue->setValue (varname, cValue (val));
        else
          queue->setValue (varname, cValue (v->asDouble() + val));
      }
    }
};

class cMacroLDec : public cMacro {
  public:
    cMacroLDec () : cMacro ("ldec") {}
    virtual void eval (const QString &params, int, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString value = params.section (' ', 1, 1, QString::SectionSkipEmpty);
      bool ok = false;
      double val = value.toDouble(&ok);
      if (ok) {
        cValue *v = queue->value (varname);
        if (!v)  // no such variable ?
          queue->setValue (varname, cValue (-val));
        else
          queue->setValue (varname, cValue (v->asDouble() - val));
      }
    }
};

class cMacroProvideRes : public cMacro {
public:
  cMacroProvideRes () : cMacro ("provide-res") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    varList(sess)->provideResource (varname);
  }
};

class cMacroRequestRes : public cMacro {
public:
  cMacroRequestRes () : cMacro ("request-res") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    bool ret = varList(sess)->requestResource (varname);
    if (!ret)
      am->invokeEvent ("message", sess, "No free resources of that type.");
  }
};

class cMacroEcho : public cMacro {
public:
  cMacroEcho () : cMacro ("echo") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
    if( !params.isEmpty() )
    {
      QString txt = expandVariables (params, sess, queue);
      am->invokeEvent ("user-message", sess, txt);
    }
  }
};

class cMacroAGroupOn : public cMacro {
public:
  cMacroAGroupOn () : cMacro ("agroupon") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    cGroupList *agl = dynamic_cast<cGroupList *>(am->object ("aliasgrouplist", sess));
    QString varname = params.stripWhiteSpace ();
    int id = agl->getGroupId (varname);
    if (id != -1)
    {
      agl->activateGroup (id, true);
      dlgGroups *dg = agl->dialog();
      if (dg) dg->updateView ();
    }
  }
};

class cMacroAGroupOff : public cMacro {
public:
  cMacroAGroupOff () : cMacro ("agroupoff") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    cGroupList *agl = dynamic_cast<cGroupList *>(am->object ("aliasgrouplist", sess));
    QString varname = params.stripWhiteSpace ();
    int id = agl->getGroupId (varname);
    if (id != -1)
    {
      agl->activateGroup (id, false);
      dlgGroups *dg = agl->dialog();
      if (dg) dg->updateView ();
    }
  }
};

class cMacroTGroupOn : public cMacro {
public:
  cMacroTGroupOn () : cMacro ("tgroupon") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    cGroupList *tgl = dynamic_cast<cGroupList *>(am->object ("triggergrouplist", sess));
    QString varname = params.stripWhiteSpace ();
    int id = tgl->getGroupId (varname);
    if (id != -1)
    {
      tgl->activateGroup (id, true);
      dlgGroups *dg = tgl->dialog();
      if (dg) dg->updateView ();
    }
  }
};

class cMacroTGroupOff : public cMacro {
public:
  cMacroTGroupOff () : cMacro ("tgroupoff") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *) {
    cGroupList *tgl = dynamic_cast<cGroupList *>(am->object ("triggergrouplist", sess));
    QString varname = params.stripWhiteSpace ();
    int id = tgl->getGroupId (varname);
    if (id != -1)
    {
      tgl->activateGroup (id, false);
      dlgGroups *dg = tgl->dialog();
      if (dg) dg->updateView ();
    }
  }
};

class cMacroTick : public cMacro {
public:
  cMacroTick () : cMacro ("tick") {}
  virtual void eval (const QString &, int sess, cCmdQueue *) {
    cTimerList *tl = dynamic_cast<cTimerList *>(am->object ("timerlist", sess));
    cTimerTick *tt = tl->getTimerTick();
    QString sTickMessage;
    if ((tt != NULL) && tt->active())
    {
      if( tt->countdown() > 1 )
        sTickMessage.sprintf( "Tick in %d seconds.", tt->countdown() );
      else if( tt->countdown() == 1 )
        sTickMessage.sprintf( "Tick in 1 second." );
      else if( tt->countdown() == 0 )
        sTickMessage.sprintf( "Tick now." );
      else
        sTickMessage.sprintf( "Resetting tick timer..." );
    }
    else
      sTickMessage = "Tick timer is not active.";
    am->invokeEvent ("message", sess, sTickMessage);
  }
};

class cMacroTickNow : public cMacro {
public:
  cMacroTickNow () : cMacro ("ticknow") {}
  virtual void eval (const QString &, int sess, cCmdQueue *) {
    cTimerList *tl = (dynamic_cast<cTimerList *>(am->object ("timerlist", sess)));
    tl->getTimerTick()->resetTickCount();
  }
};

class cMacroAddItem : public cMacro {
  public:
    cMacroAddItem () : cMacro ("additem") {}
    virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString value = params.section (' ', 1, -1, QString::SectionSkipEmpty);
      varList(sess)->valueNotEmpty(varname, queue)->addToList (value);
      // TODO: report variable change ... same for the next three macros ...
    }
};

class cMacroDelItem : public cMacro {
  public:
    cMacroDelItem () : cMacro ("delitem") {}
    virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString value = params.section (' ', 1, -1, QString::SectionSkipEmpty);
      cValue *val = varList(sess)->value(varname, queue);
      if (val) val->removeFromList (value);
    }
};

class cMacroArraySet : public cMacro {
  public:
    cMacroArraySet () : cMacro ("arrayset") {}
    virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString index = params.section (' ', 1, 1, QString::SectionSkipEmpty);
      QString value = params.section (' ', 2, -1, QString::SectionSkipEmpty);
      bool ok;
      int idx = index.toInt (&ok);
      if (!ok) return;
      varList(sess)->valueNotEmpty(varname, queue)->setItem(idx, value);
    }
};

class cMacroArrayDel : public cMacro {
  public:
    cMacroArrayDel () : cMacro ("arraydel") {}
    virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
      QString varname = params.section (' ', 0, 0, QString::SectionSkipEmpty);
      QString index = params.section (' ', 1, 1, QString::SectionSkipEmpty);
      bool ok;
      int idx = index.toInt (&ok);
      if (!ok) return;
      cValue *val = varList(sess)->value(varname, queue);
      if (val) val->removeItem (idx);
    }
};


class cMacroIf : public cMacro {
public:
  cMacroIf () : cMacro ("if") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *queue)
  {
    // we should perform matching (else preproc would fail) - true if non-zero
    QString par1 = params.section (' ', 0, 0, QString::SectionSkipEmpty);
    par1 = expandVariables (par1, sess, queue);
    bool match = (cValue(par1).asInteger () != 0);
    
    cExecStack *es = queue->execStack ("if");
    
    // first /if ? register as preproc command
    if (es->empty())
      queue->addPreprocessMacro (this);

    // put value to the exec-stack
    cExecStackItem item;
    item.setAttrib ("match", match ? 1 : 0);
    es->push (item);
  }

  virtual bool preprocess (cCmdQueue *queue, cCmdQueueEntry *qe)
  {
    cExecStack *es = queue->execStack ("if");
    
    // do nothing if the stack is empty
    if (es->empty()) return true;

    // if the top value on the stack has non-zero match, we should execute the command
    cExecStackItem item = es->top ();
    if (item.attrib ("match")) return true;
    
    // match is 0, we (probably) won't execute ...
    // we need to check for nested if-s first
    // anything else than if, else, endif - simply return false
    if (!qe->isMacro) return false;
    cExecStack *es2 = queue->execStack ("if-ignore");
    cExecStackItem item2;
    if (qe->macroName == "if")
    {
      // nested ignored if
      // put a value to if-ignore stack
      es2->push (item2);
      return false;
    }
    else if (qe->macroName == "else")
    {
      // End of if branch, execute this command, unless the if-ignore stack is non-empty.
      // If the if-ignore stack is empty, we want to exec /else.
      return es2->empty();
    }
    else if (qe->macroName == "endif")
    {
      // If we ignore some if (if-ignore stack is not empty), we pop one item from there
      // and return false. Otherwise, we return true - want to exec /endif.
      if (es2->empty())
        return true;
      es2->pop ();
      return false;
    }
    else return false;   // ignoring all other macros
  };
};

class cMacroElse : public cMacro {
public:
  cMacroElse () : cMacro ("else") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
    cExecStack *es = queue->execStack ("if");
    // display some error if the stack is empty
    if (es->empty()) {
      am->invokeEvent ("message", sess, i18n ("/else: called without the matching /if."));
      return;
    }
    // remove one if from the stack, switch its "match" value and push it back
    cExecStackItem item = es->pop ();
    item.setAttrib ("match", !(item.attrib ("match")));
    es->push (item);
  }
};

class cMacroEndIf : public cMacro {
public:
  cMacroEndIf () : cMacro ("endif") {}
  virtual void eval (const QString &params, int sess, cCmdQueue *queue) {
    cExecStack *es = queue->execStack ("if");
    // display some error if the stack is empty
    if (es->empty()) {
      am->invokeEvent ("message", sess, i18n ("/endif: called without the matching /if."));
      return;
    }
    // remove one if from the stack
    es->pop ();
    // unregister the if preprocess macro, if we're not in any /if (ie., stack empty)
    if (es->empty ())
      queue->removePreprocessMacro (cMacroManager::self()->macro ("if"));
  }
};


class cFunctionContains: public cFunction {
public:
  cFunctionContains () : cFunction ("contains") {}
  virtual cValue eval (std::list<cValue> &params, int, cCmdQueue *) {
    if (params.size() < 2) return cValue::empty();
    std::list<cValue>::iterator it = params.begin();
    cValue lst = *it;
    ++it;
    cValue val = *it;
    if (lst.isList())
      return cValue (lst.contains (val.asString()) ? 1 : 0);
    // not a list ? Doesn't contain it then !
    return cValue::empty();
  }
};

class cFunctionItem: public cFunction {
public:
  cFunctionItem () : cFunction ("item") {}
  virtual cValue eval (std::list<cValue> &params, int, cCmdQueue *) {
    if (params.size() < 2) return cValue::empty();
    std::list<cValue>::iterator it = params.begin();
    cValue lst = *it;
    ++it;
    cValue val = *it;
    if (lst.isArray())
      return cValue (lst.item (val.asInteger()));
    // not an array ? Not applicable then !
    return cValue::empty();
  }
};

class cFunctionCount: public cFunction {
  public:
    cFunctionCount () : cFunction ("count") {}
    virtual cValue eval (std::list<cValue> &params, int, cCmdQueue *) {
      int count = 0;
      std::list<cValue>::iterator it = params.begin();
      for (; it != params.end(); ++it) {
        count += (*it).size();
      }
      return cValue (count);
    }
};

class cFunctionGlobal: public cFunction {
  public:
    cFunctionGlobal () : cFunction ("global") {}
    virtual cValue eval (std::list<cValue> &params, int sess, cCmdQueue *) {
      if (params.size() == 0) return cValue::empty();
      QString varName = (*params.begin()).asString();
      cValue *val = varList(sess)->value (varName, 0);
      if (val) return *val;
      return cValue::empty();
    }
};

class cFunctionLocal: public cFunction {
  public:
    cFunctionLocal () : cFunction ("local") {}
    virtual cValue eval (std::list<cValue> &params, int, cCmdQueue *queue) {
      if ((params.size() == 0) || (!queue)) return cValue::empty();
      QString varName = (*params.begin()).asString();
      cValue *val = queue->value (varName);
      if (val) return *val;
      return cValue::empty();
    }
};

class cFunctionAttrib: public cFunction {
  public:
    cFunctionAttrib () : cFunction ("attrib") {}
    virtual cValue eval (std::list<cValue> &params, int sess, cCmdQueue *) {
      if (params.size() < 2) return cValue::empty();
      std::list<cValue>::iterator it = params.begin();
      QString object = (*it).asString();
      ++it;
      QString attrib = (*it).asString();
      return cValue (am->object (object, sess)->attrib (attrib));
    }
};

class cFunctionStrAttrib: public cFunction {
  public:
    cFunctionStrAttrib () : cFunction ("strattrib") {}
    virtual cValue eval (std::list<cValue> &params, int sess, cCmdQueue *) {
      if (params.size() < 2) return cValue::empty();
      std::list<cValue>::iterator it = params.begin();
      QString object = (*it).asString();
      ++it;
      QString attrib = (*it).asString();
      return cValue (am->object (object, sess)->strAttrib (attrib));
    }
};



/*************************************************************/
/***    registerInternalMacros, unregisterInternalMacros   ***/
/*************************************************************/


static cMacroSet *mset;
static cMacroUnset *munset;
static cMacroSetVal *msetval;
static cMacroLSet *mlset;
static cMacroLUnset *mlunset;
static cMacroInc *minc;
static cMacroDec *mdec;
static cMacroLInc *mlinc;
static cMacroLDec *mldec;
static cMacroProvideRes *mprovideres;
static cMacroRequestRes *mrequestres;
static cMacroEcho *mecho;
static cMacroAGroupOn *magroupon;
static cMacroAGroupOff *magroupoff;
static cMacroTGroupOn *mtgroupon;
static cMacroTGroupOff *mtgroupoff;
static cMacroTick *mtick;
static cMacroTickNow *mticknow;
static cMacroAddItem *madditem;
static cMacroDelItem *mdelitem;
static cMacroArraySet *marrayset;
static cMacroArrayDel *marraydel;
static cMacroIf *mif;
static cMacroElse *melse;
static cMacroEndIf *mendif;

static cFunctionContains *fcontains;
static cFunctionItem *fitem;
static cFunctionCount *fcount;
static cFunctionGlobal *fglobal;
static cFunctionLocal *flocal;
static cFunctionAttrib *fattrib;
static cFunctionStrAttrib *fstrattrib;

void registerInternalMacros ()
{
  mset = new cMacroSet;
  munset = new cMacroUnset;
  msetval = new cMacroSetVal;
  mlset = new cMacroLSet;
  mlunset = new cMacroLUnset;
  minc = new cMacroInc;
  mdec = new cMacroDec;
  mlinc = new cMacroLInc;
  mldec = new cMacroLDec;
  mprovideres = new cMacroProvideRes;
  mrequestres = new cMacroRequestRes;
  mecho = new cMacroEcho;
  magroupon = new cMacroAGroupOn;
  magroupoff = new cMacroAGroupOff;
  mtgroupon = new cMacroTGroupOn;
  mtgroupoff = new cMacroTGroupOff;
  mtick = new cMacroTick;
  mticknow = new cMacroTickNow;
  madditem = new cMacroAddItem;
  mdelitem = new cMacroDelItem;
  marrayset = new cMacroArraySet;
  marraydel = new cMacroArrayDel;
  mif = new cMacroIf;
  melse = new cMacroElse;
  mendif = new cMacroEndIf;

  fcontains = new cFunctionContains;
  fitem = new cFunctionItem;
  fcount = new cFunctionCount;
  fglobal = new cFunctionGlobal;
  flocal = new cFunctionLocal;
  fattrib = new cFunctionAttrib;
  fstrattrib = new cFunctionStrAttrib;

}

void unregisterInternalMacros ()
{
  delete mset;
  delete munset;
  delete msetval;
  delete mlset;
  delete mlunset;
  delete minc;
  delete mdec;
  delete mlinc;
  delete mldec;
  delete mprovideres;
  delete mrequestres;
  delete mecho;
  delete magroupon;
  delete magroupoff;
  delete mtgroupon;
  delete mtgroupoff;
  delete mtick;
  delete mticknow;
  delete madditem;
  delete mdelitem;
  delete marrayset;
  delete marraydel;
  delete mif;
  delete melse;
  delete mendif;

  delete fcontains;
  delete fitem;
  delete fcount;
  delete fglobal;
  delete flocal;
  delete fattrib;
  delete fstrattrib;

}

