/***************************************************************************
                    cconnprefs.cpp  -  connection preferences, aliases, ...
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : Po Jul 22 2002
    copyright            : (C) 2002-2005 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <config.h>

#include "cconnprefs.h"

#include "cactionmanager.h"
#include "cansiparser.h"
#include "cprompt.h"
#include "cstatus.h"

#include <qdir.h>
#include <kstandarddirs.h>
#include <kconfig.h>

cConnPrefs::cConnPrefs (QString name, int _sess)
  : cActionBase ("connprefs", _sess)
{
  _name = name;

  //first construct dirname
  directory = locateLocal ("appdata", "profiles/" + _name + "/");
}

cConnPrefs::~cConnPrefs()
{
  //save all preferences
  save ();

  //reset settings
  if (sess() > 0) {
    cANSIParser *ap = dynamic_cast<cANSIParser *>(object ("ansiparser"));
    if (ap) ap->setUseAnsi (true);
  }
  //IMPORTANT: we CANNOT reset telnet's quitcommand here, because
  //disconnect() (where it's needed) is called AFTER this one...
}

void cConnPrefs::load ()
{
  //prepare configfile
  KConfig *config = new KConfig (directory + "preferences", true);

  //load basic info
  config->setGroup ("General");
  setServer (config->readEntry ("Server", ""));
  setPort (config->readNumEntry ("Port", 0));
  setLogin (config->readEntry ("Login", ""));
  setPassword (config->readEntry ("Password", ""));

  //login sequence
  config->setGroup ("Login sequence");
  _connstr.clear ();
  int lsc = config->readNumEntry ("Count", 0);
  for (int i = 1; i <= lsc; i++)
  {
    QString s = config->readEntry ("Line " + QString::number(i), QString::null);
    _connstr.append (s);    
  }
  if (lsc == 0) //default login sequence
  {
    _connstr.append ("$name");
    _connstr.append ("$password");
  }
  
  //connection parameters
  config->setGroup ("Connection");
  setAnsiColors (config->readBoolEntry ("ANSI Colors", true));
  setLimitTriggers (config->readBoolEntry ("Limit triggers", true));
  setLimitRepeater (config->readBoolEntry ("Limit repeater", true));
  setNegotiateOnStartup (config->readBoolEntry ("Negotiate on startup", true));
  setPromptLabel (config->readBoolEntry ("Prompt label", false));
  setLPMudStyle (config->readBoolEntry ("LPMud style", false));
  setStatusPrompt (config->readBoolEntry ("Status prompt", false));
  setConsolePrompt (config->readBoolEntry ("Console prompt", true));
  setAutoAdvTranscript (config->readBoolEntry ("Auto logging", false));

  //commands
  config->setGroup ("Commands");
  setCmd (North, config->readEntry ("North", "n"));
  setCmd (NorthEast, config->readEntry ("NorthEast", "ne"));
  setCmd (East, config->readEntry ("East", "e"));
  setCmd (SouthEast, config->readEntry ("SouthEast", "se"));
  setCmd (South, config->readEntry ("South", "s"));
  setCmd (SouthWest, config->readEntry ("SouthWest", "sw"));
  setCmd (West, config->readEntry ("West", "w"));
  setCmd (NorthWest, config->readEntry ("NorthWest", "nw"));
  setCmd (Up, config->readEntry ("Up", "u"));
  setCmd (Down, config->readEntry ("Down", "d"));
  setQuit (config->readEntry ("Quit", "quit"));

  //scripts
  config->setGroup ("Scripts");
  setScriptDir (config->readEntry ("Script directory", QDir::homeDirPath()));
  setWorkDir (config->readEntry ("Working directory", QDir::homeDirPath()));

  //directories (the above two should go here, but we don't want to corrupt
  // existing settings)
  config->setGroup ("Directories");
  setTranscriptDir (config->readEntry ("Transcript directory", QDir::homeDirPath()));
  
#ifdef HAVE_SOUND
  //MSP
  config->setGroup ("Sound Protocol");
  sounddirs.clear ();
  setUseMSP (config->readBoolEntry ("Use MSP", true));
  setAlwaysMSP (config->readBoolEntry ("Always MSP", false));
  setMidlineMSP (config->readBoolEntry ("Midline MSP", false));
  int count = config->readNumEntry ("Path count", -1);
  if (count == -1)
  {
    //default sound dirs
    count = 1;
    sounddirs << QDir::homeDirPath() + "/sounds";
  }
  else
    for (int i = 1; i <= count; i++)
      sounddirs << config->readEntry (QString("Path ") + QString::number(i),
            QString::null);
#endif  //HAVE_SOUND

#ifdef HAVE_MXP
  //MXP
  config->setGroup ("MUD Extension Protocol");
  //how to use MXP? By default, we auto-detect it (i.e., we parse for MXP stuff, but we
  //remain in LOCKED mode by default, until changed by server)
  setUseMXP (config->readNumEntry ("Use MXP", 3));
  varprefix = config->readEntry ("Variable prefix", "");
#endif

  delete config;

}

void cConnPrefs::save ()
{
  //prepare configfile
  KConfig *config = new KConfig (directory + "preferences");

  //save basic info
  config->setGroup ("General");
  config->writeEntry ("Server", _server);
  config->writeEntry ("Port", _port);
  config->writeEntry ("Login", _login);
  config->writeEntry ("Password", _password);

  //login sequence
  config->setGroup ("Login sequence");
  int lsc = _connstr.count();
  config->writeEntry ("Count", lsc);
  QStringList::iterator it;
  int num = 1;
  for (it = _connstr.begin(); it != _connstr.end(); ++it)
  {
    config->writeEntry ("Line " + QString::number(num), *it);
    num++;
  }
  
  //connection parameters
  config->setGroup ("Connection");
  config->writeEntry ("ANSI Colors", _ansicolors);
  config->writeEntry ("Limit triggers", _limittriggers);
  config->writeEntry ("Limit repeater", _limitrepeater);
  config->writeEntry ("Negotiate on startup", _startupneg);
  config->writeEntry ("Prompt label", _promptlabel);
  config->writeEntry ("LPMud style", _lpmudstyle);
  config->writeEntry ("Status prompt", _statusprompt);
  config->writeEntry ("Console prompt", _consoleprompt);
  config->writeEntry ("Auto logging", _autoadvtranscript);

  //commands
  config->setGroup ("Commands");
  config->writeEntry ("North", _cmd[North]);
  config->writeEntry ("NorthEast", _cmd[NorthEast]);
  config->writeEntry ("East", _cmd[East]);
  config->writeEntry ("SouthEast", _cmd[SouthEast]);
  config->writeEntry ("South", _cmd[South]);
  config->writeEntry ("SouthWest", _cmd[SouthWest]);
  config->writeEntry ("West", _cmd[West]);
  config->writeEntry ("NorthWest", _cmd[NorthWest]);
  config->writeEntry ("Up", _cmd[Up]);
  config->writeEntry ("Down", _cmd[Down]);
  config->writeEntry ("Quit", _cmdquit);
  
  //scripts
  config->setGroup ("Scripts");
  config->writeEntry ("Script directory", scriptdir);
  config->writeEntry ("Working directory", defworkdir);

  //directories (the above dirs should be here as well, but we don't want to
  //corrupt settings of existing users)
  config->setGroup ("Directories");
  config->writeEntry ("Transcript directory", transcriptdir);
  
#ifdef HAVE_SOUND
  //MSP
  config->setGroup ("Sound Protocol");
  config->writeEntry ("Use MSP", usemsp);
  config->writeEntry ("Always MSP", alwaysmsp);
  config->writeEntry ("Midline MSP", midlinemsp);
  config->writeEntry ("Path count", sounddirs.count());
  int i = 0;
  for (it = sounddirs.begin(); it != sounddirs.end(); ++it)
    config->writeEntry (QString("Path ") + QString::number(++i), *it);
#endif  //HAVE_SOUND

#ifdef HAVE_MXP
  //MXP
  config->setGroup ("MUD Extension Protocol");
  //how to use MXP? By default, we auto-detect it (i.e., we parse for MXP stuff, but we
  //remain in LOCKED mode by default, until changed by server)
  config->writeEntry ("Use MXP", usemxp);
  config->writeEntry ("Variable prefix", varprefix);
#endif  //HAVE_MXP
  
  delete config;  //this saves all data
}

//some functions that get/set preferences; most of them are in header file

void cConnPrefs::setPort (int port)
{
  if ((port > 0) && (port <= 65535))
    _port = port;
}

void cConnPrefs::setAnsiColors (bool val)
{
  _ansicolors = val;
  if (sess() > 0) {
    cANSIParser *ap = dynamic_cast<cANSIParser *>(object ("ansiparser"));
    if (ap) ap->setUseAnsi (val);
  }
}

void cConnPrefs::setCmd (int which, QString val)
{
  if ((which >= 0) && (which <= 9))
    _cmd[which] = val;
}

void cConnPrefs::setQuit (QString val)
{
  _cmdquit = val;
  cActionManager::self()->callAction ("telnet", "set-quit-command", sess(), val);
}

void cConnPrefs::setPromptLabel (bool val) 
{ 
  _promptlabel = val; 
  
  if (sess() > 0) {
    cPrompt *prompt = dynamic_cast<cPrompt *>(object ("prompt"));
    cStatus *status = dynamic_cast<cStatus *>(object ("status"));
    if ((!prompt) || (!status)) return;

    if (val) {
      prompt->show ();
      status->clearPartialLine();
    }
    else {
      prompt->hide ();
    }
  }
}

void cConnPrefs::setNegotiateOnStartup (bool val)
{
  _startupneg = val;
  cActionManager::self()->callAction ("telnet", "set-negotiate-on-startup", sess(), val ? 1 : 0);
}

void cConnPrefs::setLPMudStyle(bool val)
{
  _lpmudstyle = val;
  cActionManager::self()->callAction ("telnet", "set-lp-mud-style", sess(), val ? 1 : 0);
}

void cConnPrefs::setStatusPrompt(bool val)
{
  _statusprompt = val;
  if ((!val) && (sess() > 0)) {
    cStatus *status = dynamic_cast<cStatus *>(object ("status"));
    if (status) status->clearPartialLine();
  }
}

#ifdef HAVE_MXP

void cConnPrefs::setUseMXP (int val)
{
  usemxp = val;
  cActionManager::self()->callAction ("telnet", "mxp-allowed", sess(), val, 0);
}

#endif

