//
// C++ Implementation: cMenuManager
//
// Description: Menu manager.
//
//
// Author: Tomas Mecir <kmuddy@kmuddy.net>, (C) 2006
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "cmenumanager.h"

#include <kaction.h>
#include <kmenubar.h>
#include <qpopupmenu.h>

#include <list>
#include <map>

using namespace std;

struct ItemPosition {
  QString position;
  int offset;
};

struct Position {
  QPopupMenu *menu;
  int position, count;
};

struct cMenuManagerPrivate {
  KMenuBar *menuBar;

  map<QPopupMenu *, ItemPosition> menuPositions;
  map<KAction *, ItemPosition> itemPositions;
  
  // positions; main manubar's positions have menu==0
  map<QString, Position> positions;
  list<QString> positionOrder;
};

cMenuManager *cMenuManager::_self = 0;

cMenuManager::cMenuManager () : cActionBase ("menumanager", 0)
{
  d = new cMenuManagerPrivate;
  d->menuBar = 0;
}

cMenuManager::~cMenuManager ()
{
  delete d;
}

cMenuManager *cMenuManager::self ()
{
  if (!_self) _self = new cMenuManager;
  return _self;
}

void cMenuManager::setMenuBar (KMenuBar *menuBar)
{
  d->menuBar = menuBar;
}

void cMenuManager::addMenuPosition (const QString &name)
{
  if (d->positions.count (name)) return;
  
  // compute current index of the last item
  int afterLast = 0;
  map<QString, Position>::iterator it;
  for (it = d->positions.begin(); it != d->positions.end(); ++it)
    if ((it->second.menu == 0) && ((it->second.position + it->second.count) >= afterLast))
      afterLast = it->second.position + it->second.count;
  
  // create the position
  Position pos;
  pos.menu = 0;
  pos.position = afterLast;
  pos.count = 0;
  d->positions[name] = pos;
  
  d->positionOrder.push_back (name);
}

void cMenuManager::addItemPosition (const QString &name, QPopupMenu *menu)
{
  if (d->positions.count (name)) return;
  
  // compute current index of the last item
  int afterLast = -1;
  map<QString, Position>::iterator it;
  for (it = d->positions.begin(); it != d->positions.end(); ++it)
    if ((it->second.menu == menu) && ((it->second.position + it->second.count) >= afterLast))
      afterLast = it->second.position + it->second.count;
  
  // increase afterLast by one, so it takes a separator inuto account. If there are no positions,
  // this will make afterLast = 0
  afterLast++;
  
  // create the position
  Position pos;
  pos.menu = menu;
  pos.position = afterLast;
  pos.count = 0;  // does not include separator
  d->positions[name] = pos;

  // separator gets inserted to the very end, but only if this is not the very first position
  if (afterLast)
    menu->insertSeparator ();

  d->positionOrder.push_back (name);
}

void cMenuManager::addMenu (QPopupMenu *menu, const QString &label, const QString &position,
    int id)
{
  // nothing if there's no such position
  if (!d->positions.count (position)) return;
  // nothing if it's an item position
  if (d->positions[position].menu) return;
  
  // insert the item at position
  ItemPosition pos;
  pos.position = position;
  pos.offset = d->positions[position].count;
  d->menuPositions[menu] = pos;

  // plug the actual menu 
  d->menuBar->insertItem (label, menu, id, positionOf (menu));

  // increase count for this position
  d->positions[position].count++;

  // shift the items to the right from here
  shiftItems (position, 1);
}

void cMenuManager::removeMenu (QPopupMenu *menu)
{
  if (!d->menuPositions.count (menu)) return;
  
  // get positionOf and menuOf of the item
  int pos = positionOf (menu);
  QString position = d->menuPositions[menu].position;

  // update count for position
  d->positions[position].count--;

  // actually remove item
  d->menuBar->removeItemAt (pos);

  // call shiftItems to shift backwards
  shiftItems (position, -1);
}

void cMenuManager::plug (KAction *action, QString position)
{
  // nothing if there's no such position
  if (!d->positions.count (position)) return;
  // nothing if it's a menubar position
  if (!d->positions[position].menu) return;
  
  // insert the item at position
  ItemPosition pos;
  pos.position = position;
  pos.offset = d->positions[position].count;
  d->itemPositions[action] = pos;

  // plug the actual item
  action->plug (d->positions[position].menu, positionOf (action));

  // increase count for this position
  d->positions[position].count++;

  // shift the items to the right from here
  shiftItems (position, 1);
}

void cMenuManager::unplug (KAction *action)
{
  if (!d->itemPositions.count (action)) return;
  
  // get positionOf and menuOf of the item
  QPopupMenu *menu = menuOf (action);
  QString position = d->itemPositions[action].position;

  // update count for position
  d->positions[position].count--;

  // actually remove item
  action->unplug (menu);

  // call shiftItems to shift backwards
  shiftItems (position, -1);
}

int cMenuManager::positionOf (KAction *action)
{
  if (!d->itemPositions.count (action))
    return -1;
  QString pos = d->itemPositions[action].position;
  int offset = d->itemPositions[action].offset;
  
  if (!d->positions.count (pos)) return -1;
  return d->positions[pos].position + offset;
}

int cMenuManager::positionOf (QPopupMenu *menu)
{
  if (!d->menuPositions.count (menu))
    return -1;
  QString pos = d->menuPositions[menu].position;
  int offset = d->menuPositions[menu].offset;
  
  if (!d->positions.count (pos)) return -1;
  return d->positions[pos].position + offset;
}

QPopupMenu *cMenuManager::menuOf (KAction *action)
{
  if (!d->itemPositions.count (action))
    return 0;
  QString pos = d->itemPositions[action].position;
  
  if (!d->positions.count (pos)) return 0;
  return d->positions[pos].menu;
}

void cMenuManager::shiftItems (const QString &position, int delta)
{  
  // nothing if there's no such position
  if (!d->positions.count (position)) return;
  QPopupMenu *menu = d->positions[position].menu;
  int startFrom = d->positions[position].position;

  // positions are traversed in the order of creation, which means that positions with lower
  // index are first in positionOrder - we start moving only after we hit the current position
  bool gotPast = false;

  // walk through position definitions and update those that need updating
  list<QString>::iterator it;
  for (it = d->positionOrder.begin(); it != d->positionOrder.end(); ++it)
  {
    if (!gotPast) {
      if (*it == position) gotPast = true;
      continue;
    }
    if ((d->positions[*it].menu == menu) && (d->positions[*it].position >= startFrom))
      d->positions[*it].position += delta;
  }
}

