/***************************************************************************
                          csaveablefield.cpp  -  alias/trigger baseclass
    This file is a part of KMuddy distribution.
                             -------------------
    begin                : So sep 7 2002
    copyright            : (C) 2002 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "csaveablefield.h"

#include "cexpresolver.h"
#include "cgroup.h"
#include "cgrouplist.h"
#include "ccmdqueue.h"
#include "cmacromanager.h"
#include "cvalue.h"

#include <kconfig.h>
#include <qregexp.h>

cSaveableField::cSaveableField ()
{
  prev = NULL;
  next = NULL;
  text = "";
  _marked = false;
  _cs = true;
  t_regexp.setCaseSensitive (true);
  backreflist.clear ();
  backrefpos = 0;
  lastlen = lastpos = 0;
  prefix = suffix = lasttext = "";
  exp = 0;
  nogroup = false;
  ourgroup = 0;
  globalmatch = false;
  matchingPos = 0;
  _group = 0;
}

cSaveableField::~cSaveableField ()
{
  backreflist.clear ();
}

bool cSaveableField::match (const QString &what)
{
  //do NOTHING if my text has zero length...
  if (text.length() == 0)
    return false;

  //no matching if the group is not activated
  if ((!nogroup) && (ourgroup != NULL) && (!ourgroup->active()))
    return false;

  if (backrefpos != 0)
    delete[] backrefpos;
  backrefpos = 0;
  
  bool matched = false;
  int n;
  
  switch (type) {
    case exact:
      if (matchingPos != 0)
        matched = false;
      else if (_cs)
        matched = (what == text);
      else
        matched = (what.lower() == text.lower());
      if (matched)
      {
        prefix = suffix = "";
        lastpos = 0;
        lastlen = text.length ();
      }
    break;
    case substring:
      n = what.find (text, matchingPos, _cs);
      matched = (n != -1);
      if (matched)
      {
        prefix = what.left (n);
        suffix = what.right (what.length() - (n + text.length()));
        lastpos = n;
        lastlen = text.length ();
      }
    break;
    case begin:
      if (matchingPos != 0)
        matched = false;
      else if (_cs)
        matched = what.startsWith (text);
      else
        matched = what.lower().startsWith (text.lower());
      if (matched)
      {
        prefix = "";
        suffix = what.right (what.length() - text.length());
        lastpos = 0;
        lastlen = text.length ();
      }      
    break;
    case end:
      if (matchingPos != 0)
        matched = false;
      else if (_cs)
        matched = what.endsWith (text);
      else
        matched = what.lower().endsWith (text.lower());
      if (matched)
      {
        prefix = what.left (what.length() - text.length());
        suffix = "";
        lastpos = what.length() - text.length();
        lastlen = text.length ();
      }
    break;
    case regexp:
      //regexp's case-sensitivity is set in constructor and in function
      // setCaseSensitive
      n = t_regexp.search (what, matchingPos);
      if (n != -1)    //MATCH!
      {
        matched = true;
        lastpos = n;
        lastlen = t_regexp.matchedLength();
        prefix = what.left (n);
        suffix = what.right (what.length() - (n + lastlen));
        backreflist.clear ();
        backreflist = t_regexp.capturedTexts ();
        //positions of back-references
        int npos = backreflist.count();
        backrefpos = new int[npos];
        for (int i = 0; i < npos; i++)
          backrefpos[i] = t_regexp.pos (i);
      }
    break;
  };

  //return the result...
  if (matched)
  {
    lasttext = what;
    return true;
  }
  else
    return false;
}

bool cSaveableField::testCondition (int sess, cExpResolver *resolver)
{
  // no condition -> always matches ...
  if (!exp) return true;
  
  cCmdQueue *queue = new cCmdQueue (sess);
  resolver->setQueue (queue);
  queue->setValue ("prefix", prefix);
  queue->setValue ("suffix", suffix);
  queue->setValue ("prefixtrim", prefix.stripWhiteSpace());
  queue->setValue ("suffixtrim", suffix.stripWhiteSpace());
  queue->setValue ("matched", lasttext.mid (lastpos, lastlen));
  if (type == regexp)
    for (unsigned int i = 0; i < backreflist.count(); ++i)
      queue->setValue (QString::number (i), backreflist[i]);
  
  cValue val = exp->evaluate (resolver);
  delete queue;
  resolver->setQueue(0);

  // test passes if the evaluator returns non-zero ...
  return (val.asInteger() != 0);
}

void cSaveableField::setText (const QString &_text)
{
   text = _text;

   if (type == regexp)
   {
      t_regexp.setPattern(text); // Parse regular expression
   } // endif need to parse regular expression
}

// Pseudo-variables could in theory be done using local variables in a command queue,
// the problem is that aliases need them too, thus we cannot expand like that ...
// Hence we need to have this thingie ...
void cSaveableField::expandPseudoVariables (QString &string)
{
  QString newstring = "";
  int len = string.length ();
  bool invar = false;
  bool inpar = false; //variable name is in parentheses '(' and ')'
  QString varname;
  for (int i = 0; i < len; i++)
  {
    QChar ch = string[i];

    //support things like $$$a correctly (when mixing pseudo and normal vars)
    if (invar && (!inpar) && (ch == '$') && (varname.isEmpty()))
    {
      invar = false;
      newstring += ch;
    }

    if (!invar)
    {
      if (ch == '$')
      {
        invar = true;
        inpar = false;
        varname = "";
      }
      else
        newstring += ch;
    }
    else
    {
      if ((varname == "") && (ch == '(') && (!inpar))
        inpar = true;
      else
      if (!(ch.isLetterOrNumber ()))  //end of pseudo-variable name
      {
        invar = false;

        if (inpar)
        {
          inpar = false;
          if (ch == ')')
          {
            QString newstr = processVariable (varname);
            if (newstr == QString::null)
              newstring += "$(" + varname + ")";
            else
              newstring += newstr;
          }
          else
            newstring += "$(" + varname + ((ch == '$') ? QString::null : QString(ch));
        }
        else
        {
          QString newstr = processVariable (varname);
          if (newstr == QString::null)
            newstring += "$" + varname;
          else
            newstring += newstr;
          if (ch != '$')
            newstring += ch;
        }
        if (ch == '$')  //new variable follows immediately
        {
          invar = true;
          varname = "";
        }
      }
      else
        varname += ch;
    }
  }

  if (invar)    //if a variable ends the string
    if (inpar)
      newstring += "$(" + varname;  //no ending parenthesis - no variable
    else
    {
      QString newstr = processVariable (varname);
      newstring += (newstr == QString::null) ? ("$"+varname) : newstr;
    }
  
  //apply changes!
  string = newstring;
}

void cSaveableField::setCond (const QString &c)
{
  condition = c;
  delete exp;
  exp = 0;
  
  // no expression ? nothing to do !
  if (c.stripWhiteSpace().isEmpty()) return;

  // parse the condition
  exp = new arith_exp;
  bool ok = exp->compile (condition);
  if (!ok) {
    // cannot parse condition - no conditional matching ...
    delete exp;
    exp = 0;
  }
}

QString cSaveableField::processVariable (const QString &varname)
{
  //look if we know the name; return the real string if we do...
  QString result;
  bool ok;
  int number = varname.toInt (&ok);
  if (ok)   //it was a number
  {
    if (type == regexp)
    {
      if (number >= (int) backreflist.count())
        result = QString::null;
      else
        result = backreflist[number];
    }
    else
      result = QString::null;
  }
  else
  {
    if (varname == "prefix")
      result = prefix;
    else
    if (varname == "suffix")
      result = suffix;
    else
    if (varname == "prefixtrim")
      result = prefix.stripWhiteSpace ();
    else
    if (varname == "suffixtrim")
      result = suffix.stripWhiteSpace ();
    else
    if (varname == "matched")
      result = lasttext.mid (lastpos, lastlen);
    else
      result = QString::null;
  }
  return result;
}

void cSaveableField::setType (int t)
{
  if ((t >= 0) && (t < TYPES))
    type = t;

  if (type == regexp)
  {
    t_regexp.setPattern(text); // Parse regular expression
  } // endif regular expression
}

void cSaveableField::setCaseSensitive (bool cs)
{
  _cs = cs;
  //set case sensitivity for regexp
  t_regexp.setCaseSensitive (cs);
}

void cSaveableField::updateGroupPointer (cGroupList *list)
{
  ourgroup = 0;
  cGroup *basegroup = 0;
  if (nogroup)
    return;
  //find our group in the list of groups...
  for (list->reset(); *list; (*list)++)
  {
    cSaveableField *field = **list;
    cGroup *g = (cGroup *) field;
    if (g->number() == 0) //this is BASE group
      basegroup = g;
    if (g->number() == _group) //found!
    {
      ourgroup = g;
      break;
    }
  }
  if (ourgroup == 0)  //if there is no group with such number
  {
    _group = 0;           //then we put ourselves into the BASE group
    ourgroup = basegroup; // (cGroupList guarantees that BASE group always exists)
  }
}


