/***************************************************************************
                          cvariablelist.cpp  -  list of variables
                             -------------------
    begin                : Po sep 8 2003
    copyright            : (C) 2003 by Tomas Mecir
    email                : kmuddy@kmuddy.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cvariablelist.h"

#include "ccmdqueue.h"
#include "cvalue.h"
#include "cvariable.h"

#include <kconfig.h>

using namespace std;

//maximum number of passes when doing recursive variable expansion

#define MAXEXPANDPASSES 10

cVariableList::cVariableList (QString filename, int sess)
      : cSaveableList ("variablelist", sess)
{
  prototype = new cVariable;
  objName = "Variable";
  config = new KConfig (filename);
  load ();
  initMapping ();
}

cVariableList::~cVariableList ()
{
  save ();
  config->sync ();
  vars.clear ();
}

void cVariableList::initMapping ()
{
  //traverse the list of variables, adding each of them to the mapping
  for (reset(); *this; (*this)++)
  {
    cVariable *var = (cVariable *) cur;
    //no duplicity check - assumes that the savefile is sane
    //consequence: don't mess up the variable savefile!
    vars[var->name()] = var;
  }
}

QString cVariableList::getValue (const QString &varname)
{
  cValue *val = value (varname);
  //no such variable -> nothing to do
  if (!val) return QString::null;

  QString v = val->asString();
  //we mustn't return QString::null if the variable contains it, because
  //that would be treated as if the variable hadn't existed; return empty
  //string instead
  if (v == QString::null)
    return QString ("");
  else
    return v;
}

int cVariableList::getIntValue (const QString &varname)
{
  cValue *val = value (varname);
  //no such variable -> nothing to do
  if (!val) return 0;

  return val->asInteger();
}

cValue *cVariableList::value (const QString &varname, cCmdQueue *queue)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  if (queue) {
    cValue *value = queue->value (vn);
    if (value)
      return value;
  }

  if (!exists (varname))
    return 0;
  
  return vars[vn]->getValue();
}

cValue *cVariableList::valueNotEmpty (const QString &varname, cCmdQueue *queue)
{
  cValue *val = value (varname, queue);
  if (val) return val;
  // doesn't exist - create empty !
  val = new cValue;
  set (varname, val);
  return val;
}
  
bool cVariableList::exists (const QString &varname)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);
  return (vars.count (vn) != 0);
}

void cVariableList::set (const QString &varname, const QString &value)
{
  cValue *v = new cValue (value);
  set (varname, v);
  delete v;
}

void cVariableList::set (const QString &varname, cValue *value)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  QString oldval = QString::null;
  if (exists (vn))
  {
    //variable already exists - we just update its value
    oldval = vars[vn]->value ();
    vars[vn]->setValue (value);
  }
  else
  {
    //good, add new variable to the list
    cVariable *var = new cVariable;
    var->setName (vn);
    var->setValue (value);
    addToEnd (var);
    vars[vn] = var;
  }

  //inform everyone about the change ...
  invokeEvent ("old-var-value", sess(), vn, oldval);
  invokeEvent ("var-changed", sess(), vn, value ? value->asString() : QString::null);
}

void cVariableList::unset (const QString &varname)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  //no such variable -> nothing to do
  if (!exists (vn))
    return;

  //get the variable
  cVariable *var = vars[vn];
  //update the mapping
  vars.erase (vn);
  //update the list (also frees memory)
  cur = var;
  removeCurrent ();
}

void cVariableList::inc (const QString &varname, double delta)
{
  if (!exists (varname)) {
    cValue val (delta);
    set (varname, &val);
  } else {
    QString vn = varname;
    if (varname[0] == '$')
      vn = vn.mid(1);
    cValue *val = vars[vn]->getValue();
    if (val) {
      QString oldval = val->asString();
      val->setValue (val->asDouble() + delta);
      // we are bypassing cVariableList::set here, hence we need to invoke
      // this event ourselves ...
      invokeEvent ("old-var-value", sess(), vn, oldval);
      invokeEvent ("var-changed", sess(), vn, val->asString());
    }
    else {
      cValue d (delta);
      set (varname, &d);
    }
  }
}

void cVariableList::dec (const QString &varname, double delta)
{
  inc (varname, -1.0 * delta);
}

void cVariableList::provideResource (const QString &varname)
{
  //returns a resource to available ones
  inc (varname, 1);
}

bool cVariableList::requestResource (const QString &varname)
{
  //is a resource available?
  if (getIntValue (varname) <= 0)
    //no :(
    return false;
  //yes, "allocate" one
  dec (varname, 1);
  return true;
}

QStringList cVariableList::getList ()
{
  map<QString, cVariable *>::iterator it;
  QStringList list;
  for (it = vars.begin(); it != vars.end(); ++it)
    list.push_back (it->first);
  return list;
}

QString cVariableList::expandVariables (const QString &string, bool recursive, cCmdQueue *queue)
{
  return doExpandVariables (string, recursive ? MAXEXPANDPASSES : 1, queue);
}

QString cVariableList::doExpandVariables (const QString &string, int recursionCounter,
    cCmdQueue *queue)
{
  //almost identical with cSaveableField::expandPseudoVariables

  QString newstring = "";
  int len = string.length ();
  bool hadvar = false; //whether we had some variable - used to prevent
                       //unneeded recursion
  bool invar = false;
  bool inpar = false; //variable name is in parentheses '(' and ')'
  QString varname;
  for (int i = 0; i < len; i++)
  {
    QChar ch = string[i];

    //support things like $$$a correctly
    if (invar && (!inpar) && (ch == '$') && (varname.isEmpty()))
    {
      invar = false;
      newstring += ch;
    }

    if (!invar)
    {
      if (ch == '$')
      {
        invar = true;
        inpar = false;
        varname = "";
      }
      else
        newstring += ch;
    }
    else
    {
      if ((varname == "") && (ch == '(') && (!inpar))
        inpar = true;
      else
      if ((!ch.isLetterOrNumber())&&
          (ch != '_')               )  //end of variable name
      {
        invar = false;

        if (inpar)
        {
          inpar = false;
          if (ch == ')')
          {
            QString newstr = processVariable (varname, queue);
            if (newstr == QString::null)
              newstring += "$(" + varname + ")";
            else
            {
              newstring += newstr;
              hadvar = true;
            }
          }
          else
            newstring += "$(" + varname + ch;
        }
        else
        {
          QString newstr = processVariable (varname, queue);
          if (newstr == QString::null)
            newstring += "$" + varname;
          else
          {
            newstring += newstr;
            hadvar = true;  
          }
          newstring += ch;
        }
      }
      else
        varname += ch;
    }
  }

  if (invar)    //if a variable ends the string
    if (inpar)
      newstring += "$(" + varname;  //no ending parenthesis - no variable
    else
    {
      QString newstr = processVariable (varname, queue);
      if (newstr == QString::null)
        newstring += "$" + varname;
      else
      {
        newstring += newstr;
        hadvar = true;
      }
    }

  //another pass if needed, then return the result
  recursionCounter--;
  if (hadvar && (recursionCounter > 0))
    return doExpandVariables (newstring, recursionCounter, queue);
  return newstring;
}

QString cVariableList::processVariable (const QString &varname, cCmdQueue *queue)
{
  QString vn = varname;
  if (varname[0] == '$')
    vn = vn.mid(1);

  if (queue) {
    QString localValue = queue->getValue (vn);
    if (localValue != QString::null)
      return localValue;
  }
  return getValue (vn);
}

