//
// C++ Implementation: cnotifymanager
//
// Description: Notification manager for the /notify macro
//
//
// Author: Tomas Mecir <kmuddy@kmuddy.net>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "cnotifymanager.h"

#include <kextsock.h>

#include <iostream>

using namespace std;

cNotifyManager::cNotifyManager ()
{
  notifications.clear();
}

cNotifyManager::~cNotifyManager()
{
}

/** send information to external IP port */
void cNotifyManager::doNotify(const int  port,
                             QString   &ip_data)
{
  std::list <cEventNotification *>::iterator ptr  = notifications.begin();
  std::list <cEventNotification *>::iterator last = notifications.end();

  while ((ptr                != last)&&
          ((*ptr)->port_num() != port)  )
  {
    ++ptr;
  } // endwhile more 'in progress' event notifications to search

  if (ptr == last)
  {
    cEventNotification *notify_event = new cEventNotification(port,ip_data);

    notifications.push_back(notify_event);

    connect(notify_event,
            SIGNAL (finished(cEventNotification *)),
            this,
            SLOT (portNotified(cEventNotification *)));

    notify_event->send();

  } // endif no other notifications in progress to this IP port at this time
  else
  {
    // Queue data for sending later
    (*ptr)->queue(ip_data);

  } // endelse need to queue notification

}


void cNotifyManager::portNotified(cEventNotification *event_notify)
{
  std::list<cEventNotification *>::iterator ptr  = notifications.begin();
  std::list<cEventNotification *>::iterator last = notifications.end();

  while ((ptr  != last)        &&
          (*ptr != event_notify)  )
  {
    ++ptr;
  } // endwhile event notification not found

  if (ptr != last)
  {
    disconnect(*ptr,
                SIGNAL (finished(cEventNotification *)),
                this,
                SLOT (portNotified(cEventNotification *)));

    delete *ptr;

    notifications.erase(ptr);
  } // endif found relevant event notification

}


// cEventNotification implementation from here on

//***************************************************************
// Constructor - set up signals and initialise variables
//***************************************************************

cEventNotification::cEventNotification(const int      ip_port,
                                       const QString &ip_data)
{
  QString data;

  port = ip_port;
  data = ip_data+"\n";

  data_list.push_back(data);

  // Set up signal handling

  sock = new KExtendedSocket();

  connect(sock, SIGNAL(connectionSuccess()), this, SLOT(connected()));
  connect(sock, SIGNAL(connectionFailed(int)), this, SLOT(error()));
  connect(sock, SIGNAL(readyWrite()), this, SLOT(wroteAll()));
  connect(sock, SIGNAL(closed(int)), this, SLOT (connectionClosed()));

  sock->setSocketFlags(KExtendedSocket::inetSocket |
      KExtendedSocket::bufferedSocket |
      KExtendedSocket::streamSocket |
      KExtendedSocket::noResolve);

} // cEventNotification::cEventNotification


//***************************************************************
// Destructor
//***************************************************************

cEventNotification::~cEventNotification()
{
  if (sock != NULL)
  {
    disconnect(sock, SIGNAL(connectionSuccess()), this, SLOT(connected()));
    disconnect(sock, SIGNAL(connectionFailed(int)), this, SLOT(error()));
    disconnect(sock, SIGNAL(readyWrite()), this, SLOT(wroteAll()));
    disconnect(sock, SIGNAL(closed(int)), this, SLOT (connectionClosed()));

    sock->cancelAsyncConnect();

      // delete socket
    sock->deleteLater();
    sock = NULL;
  } // endif socket active

} // cEventNotification::~cEventNotification


//***************************************************************
// Function to initiate information sending to the specified IP port
//***************************************************************

void cEventNotification::send()
{
  QString dest("127.0.0.1");

  sock->setAddress(dest,port);

   // Give up on connecting after 3 seconds
  sock->setTimeout(3);

  if (sock->startAsyncConnect() == -1)
  {
    cerr << "/notify error: Unable to connect\n";

    error();
  } // endif unable to commence asynchronous connect

} // cEventNotification::send


//***************************************************************
// Queue data for sending at next opportunity
//***************************************************************

void cEventNotification::queue(const QString &ip_data)
{
//  cerr << "Queueing to send: " << ip_data.latin1() << endl;
//  cerr.flush();

  QString data = ip_data+"\n";

  data_list.push_back(data);

  if (data_list.size() == 1)
  {
    std::list<QString>::iterator ptr = data_list.begin();

      // Note: this might not buffer the entire message contents.  Not completely sure
    int written = sock->writeBlock(ptr->latin1(),
                                   ptr->length());

    if (written != (int)ptr->length())
    {
         // NOTE: Far as I can tell, this never happens.  It's just here in case it ever does
      cerr << "/notify warning: writeBlock() was unable to buffer all data\n";
    } // endif unable to buffer all requested data

    if (written == -1)
    {
        // NOTE: This shouldn't ever happen because we wait until the buffer is empty before
        // trying to send any more data, but just in case...
      cerr << "/notify warning: writeBlock() buffer is full\n";
      data_list.clear();

      connectionClosed();
    } // endif unable to write data (write buffer is full)

      // turn on signal emission
    sock->enableWrite(true);
  } // endif queue was empty - need to restart sending

} // cEventNotification::queue


//***************************************************************
// Called when connected to IP port successfully
//***************************************************************

void cEventNotification::connected()
{
  std::list<QString>::iterator ptr = data_list.begin();

   // enable buffering
  sock->setBufferSize(-1,   // Unlimited read buffer size
                       -1);  // Unlimited write buffer size

   // Note: this might not buffer the entire message contents.  Not completely sure
  int written = sock->writeBlock(ptr->latin1(),
                                 ptr->length());

  if (written != (int)ptr->length())
  {
      // NOTE: Far as I can tell, this never happens.  It's just here in case it ever does
    cerr << "/notify warning: writeBlock() was unable to buffer all data\n";
  } // endif unable to buffer all requested data

  if (written == -1)
  {
     // NOTE: This shouldn't ever happen because we wait until the buffer is empty before
     // trying to send any more data, but just in case...
    cerr << "/notify warning: writeBlock() buffer is full\n";
    data_list.clear();

    connectionClosed();
  } // endif unable to write data (write buffer is full)

   // turn on signal emission
  sock->enableWrite(true);

} // cEventNotification::connected



//***************************************************************
// Wrote some data to the IP port
//***************************************************************

void cEventNotification::wroteAll()
{
   // At this point, the KExtendedSocket write buffer is empty

  data_list.pop_front();

  if (data_list.size() == 0)
  {
      // turn off signal emission
    sock->enableWrite(false);

      // TODO: Set up some kind of tick timer to close connection after a period of inactivity?
      // NOTE: doesn't seem to be necessary because the next attempt to send to an port that's been
      //       closed will give an error, thus causing a buffer flush and the socket to be closed.

  } // endif no more data to send at the moment
  else
  {
    std::list<QString>::iterator ptr = data_list.begin();

    int written = sock->writeBlock(ptr->latin1(),
                                   ptr->length());

    if (written != (int)ptr->length())
    {
         // NOTE: Far as I can tell, this never happens.  It's just here in case it ever does
      cerr << "/notify warning: writeBlock() was unable to buffer all data\n";
    } // endif unable to buffer all requested data

    if (written == -1)
    {
        // NOTE: This shouldn't ever happen because we wait until the buffer is empty before
        // trying to send any more data, but just in case...
      cerr << "/notify warning: writeBlock() buffer is full\n";
      data_list.clear();

      connectionClosed();
    } // endif unable to write data (write buffer is full)

  } // endelse more data to send

} // cEventNotification::wroteAll



//***************************************************************
// Connection closed - all done
//***************************************************************

void cEventNotification::connectionClosed()
{
   // Empty notification queue
  data_list.clear();

   // turn off signal emission
  sock->enableWrite(false);

  disconnect(sock, SIGNAL(connectionSuccess()), this, SLOT(connected()));
  disconnect(sock, SIGNAL(connectionFailed(int)), this, SLOT(error()));
  disconnect(sock, SIGNAL(readyWrite()), this, SLOT(wroteAll()));
  disconnect(sock, SIGNAL(closed(int)), this, SLOT (connectionClosed()));

  sock->cancelAsyncConnect();

   // delete socket
  sock->deleteLater();
  sock = NULL;

  if (port != -1)
  {
    port = -1;
    emit finished(this);
  } // endif not already announced we're finished

} // cEventNotification::connectionClosed


//***************************************************************
// An error occured
//***************************************************************

void cEventNotification::error()
{
   // Usually called when we attempt to connect to a port that's no longer
   // got a server attached to it.

  data_list.clear();

  if (port != -1)
  {
    port = -1;

    emit finished(this);

  } // endif not already announced we're finished

} // cEventNotification::error



#include "cnotifymanager.moc"
